from __future__ import annotations

import copy
import linecache
import logging
from io import StringIO
from typing import TYPE_CHECKING

import numpy as np
from monty.io import zopen
from monty.json import MSONable

from pymatgen.io.pwmat.inputs import ACstrExtractor, AtomConfig, LineLocator

if TYPE_CHECKING:
    from pymatgen.core import Structure
    from pymatgen.util.typing import PathLike

logger = logging.getLogger(__name__)

__author__ = "Hanyu Liu"
__email__ = "domainofbuaa@gmail.com"
__date__ = "2024-1-16"


class Movement(MSONable):
    """Parser for data in MOVEMENT which records trajectory during MD."""

    def __init__(
        self,
        filename: PathLike,
        ionic_step_skip: int | None = None,
        ionic_step_offset: int | None = None,
    ):
        """Initialization function.

        Args:
            filename (PathLike): The path of MOVEMENT
            ionic_step_skip (int | None, optional): If ionic_step_skip is a number > 1,
                only every ionic_step_skip ionic steps will be read for
                structure and energies. This is very useful if you are parsing
                very large MOVEMENT files. Defaults to None.
            ionic_step_offset (int | None, optional): Used together with ionic_step_skip.
                If set, the first ionic step read will be offset by the amount of
                ionic_step_offset. Defaults to None.
        """
        self.filename: PathLike = filename
        self.ionic_step_skip: int | None = ionic_step_skip
        self.ionic_step_offset: int | None = ionic_step_offset
        self.split_mark: str = "--------------------------------------"

        self.chunk_sizes, self.chunk_starts = self._get_chunk_info()
        self.n_ionic_steps: int = len(self.chunk_sizes)

        self.ionic_steps: list[dict] = self._parse_sefv()
        if self.ionic_step_offset and self.ionic_step_skip:
            self.ionic_steps = self.ionic_steps[self.ionic_step_offset :: self.ionic_step_skip]

    def _get_chunk_info(self) -> tuple[list[int], list[int]]:
        """Split MOVEMENT into many chunks, so that program process it chunk by chunk.

        Returns:
            tuple[list[int], list[int]]:
                chunk_sizes (list[int]): The number of lines occupied by structural
                    information in each step.
                chunk_starts (list[int]): The starting line number for structural
                    information in each step.
        """
        chunk_sizes: list[int] = []
        row_idxs: list[int] = LineLocator.locate_all_lines(self.filename, self.split_mark)
        chunk_sizes.append(row_idxs[0])
        chunk_sizes.extend(row_idxs[ii] - row_idxs[ii - 1] for ii in range(1, len(row_idxs)))
        chunk_sizes_bak: list[int] = copy.deepcopy(chunk_sizes)
        chunk_sizes_bak.insert(0, 0)
        chunk_starts: list[int] = np.cumsum(chunk_sizes_bak).tolist()
        chunk_starts.pop(-1)
        return chunk_sizes, chunk_starts

    @property
    def atom_configs(self) -> list[Structure]:
        """AtomConfig for structures contained in MOVEMENT file.

        Returns:
            list[Structure]: List of Structure objects for the structure at each ionic step.
        """
        return [step["atom_config"] for step in self.ionic_steps]

    @property
    def e_tots(self) -> np.ndarray:
        """Total energies of each ionic step structures contained in MOVEMENT.

        Returns:
            np.ndarray: Total energy of of each ionic step structure,
                with shape of (n_ionic_steps,).
        """
        return np.array([step["e_tot"] for step in self.ionic_steps])

    @property
    def atom_forces(self) -> np.ndarray:
        """Forces on atoms in each structures contained in MOVEMENT.

        Returns:
            np.ndarray: The forces on atoms of each ionic step structure,
                with shape of (n_ionic_steps, n_atoms, 3).
        """
        return np.array([step["atom_forces"] for step in self.ionic_steps])

    @property
    def e_atoms(self) -> np.ndarray:
        """Individual energies of atoms in each ionic step structures
        contained in MOVEMENT.

        Returns:
            np.ndarray: The individual energy of atoms in each ionic step structure,
                with shape of (n_ionic_steps, n_atoms).
        """
        return np.array([step["eatoms"] for step in self.ionic_steps if ("eatoms" in step)])

    @property
    def virials(self) -> np.ndarray:
        """Virial tensor of each ionic step structure contained in MOVEMENT.

        Returns:
            np.ndarray: The virial tensor of each ionic step structure,
                with shape of (n_ionic_steps, 3, 3)
        """
        return np.array([step["virial"] for step in self.ionic_steps if ("virial" in step)])

    def _parse_sefv(self) -> list[dict]:
        """Parse the MOVEMENT file, return information ionic step structure containing
        structures, energies, forces on atoms and virial tensor.

        Returns:
            list[dict]: Structure containing structures, energies, forces on atoms
                and virial tensor. The corresponding keys are 'atom_config', 'e_tot',
                'atom_forces' and 'virial'.
        """
        ionic_steps: list[dict] = []
        with zopen(self.filename, mode="rt", encoding="utf-8") as mvt:
            tmp_step: dict = {}
            for ii in range(self.n_ionic_steps):
                tmp_chunk: str = ""
                for _ in range(self.chunk_sizes[ii]):
                    tmp_chunk += mvt.readline()
                tmp_step["atom_config"] = AtomConfig.from_str(tmp_chunk)
                tmp_step["e_tot"] = ACstrExtractor(tmp_chunk).get_e_tot()[0]
                tmp_step["atom_forces"] = ACstrExtractor(tmp_chunk).get_atom_forces().reshape(-1, 3)
                e_atoms: np.ndarray | None = ACstrExtractor(tmp_chunk).get_atom_forces()
                if e_atoms is not None:
                    tmp_step["atom_energies"] = ACstrExtractor(tmp_chunk).get_atom_energies()
                else:
                    logger.info(f"Ionic step #{ii} : Energy deposition is turn down.")
                virial: np.ndarray | None = ACstrExtractor(tmp_chunk).get_virial()
                if virial is not None:
                    tmp_step["virial"] = virial.reshape(3, 3)
                else:
                    logger.info(f"Ionic step #{ii} : No virial information.")
                ionic_steps.append(tmp_step)
        return ionic_steps


class OutFermi(MSONable):
    """Extract fermi energy (eV) from OUT.FERMI."""

    def __init__(self, filename: PathLike):
        """Initialization function.

        Args:
            filename (PathLike): The absolute path of OUT.FERMI file.
        """
        self.filename: PathLike = filename
        with zopen(self.filename, mode="rt", encoding="utf-8") as file:
            self._e_fermi: float = np.round(float(file.readline().split()[-2].strip()), 3)

    @property
    def e_fermi(self) -> float:
        """The fermi energy level.

        Returns:
            float: Fermi energy level.
        """
        return self._e_fermi


class Report(MSONable):
    """Extract information of spin, kpoints, bands, eigenvalues from REPORT file."""

    def __init__(self, filename: PathLike):
        """Initialization function.

        Args:
            filename (PathLike): The absolute path of REPORT file.
        """
        self.filename: PathLike = filename
        self._spin, self._num_kpts, self._num_bands = self._parse_band()
        self._eigenvalues = self._parse_eigen()
        self._kpts, self._kpts_weight, self._hsps = self._parse_kpt()

    def _parse_band(self) -> tuple[int, int, int]:
        """Parse REPORT file to obtain spin switches, the number of kpoints
        and the number of bands.

        Returns:
            tuple[int, int, int]: containing:
                spin (int): Whether turn on spin or not
                    1: turn down the spin
                    2: turn on the spin.
                num_kpts (int): The number of kpoints.
                num_bands (int): The number of bands.
        """
        content: str = "SPIN"
        row_idx: int = LineLocator.locate_all_lines(file_path=self.filename, content=content)[0]
        spin = int(linecache.getline(str(self.filename), row_idx).split()[-1].strip())

        content = "NUM_KPT"
        row_idx = LineLocator.locate_all_lines(file_path=self.filename, content=content)[0]
        num_kpts = int(linecache.getline(str(self.filename), row_idx).split()[-1].strip())

        content = "NUM_BAND"
        row_idx = LineLocator.locate_all_lines(file_path=self.filename, content=content)[0]
        num_bands = int(linecache.getline(str(self.filename), row_idx).split()[-1].strip())
        return spin, num_kpts, num_bands

    def _parse_eigen(self) -> np.ndarray:
        """Parse REPORT file to obtain information about eigenvalues.

        Returns:
            np.ndarray: Eigenvalues with shape of (1 or 2, n_kpoints, n_bands).
                The first index represents spin, the second index represents
                kpoints, the third index represents band.
        """
        num_rows: int = int(np.ceil(self._num_bands / 5))
        content: str = "eigen energies, in eV"
        rows_lst: list[int] = LineLocator.locate_all_lines(file_path=self.filename, content=content)
        rows_array: np.ndarray = np.array(rows_lst).reshape(self._spin, -1)
        eigenvalues: np.ndarray = np.zeros((self._spin, self._num_kpts, self._num_bands))

        for ii in range(self._spin):
            for jj in range(self._num_kpts):
                tmp_eigenvalues_str = ""
                for kk in range(num_rows):
                    tmp_eigenvalues_str += linecache.getline(str(self.filename), rows_array[ii][jj] + kk + 1)
                tmp_eigenvalues_array = np.array([float(eigen_value) for eigen_value in tmp_eigenvalues_str.split()])
                for kk in range(self._num_bands):
                    eigenvalues[ii][jj][kk] = tmp_eigenvalues_array[kk]
        return eigenvalues

    def _parse_kpt(self) -> tuple[np.ndarray, np.ndarray, dict[str, np.ndarray]]:
        """Parse REPORT file to obtain information about kpoints.

        Returns:
            3-tuple containing:
                kpts (np.ndarray): The fractional coordinates of kpoints.
                kpts_weight (np.ndarray): The weight of kpoints.
                hsps (dict[str, np.ndarray]): The name and coordinates of high symmetric points.
        """
        num_rows: int = int(self._num_kpts)
        content: str = "total number of K-point:"
        row_idx: int = LineLocator.locate_all_lines(self.filename, content)[0]
        kpts: np.ndarray = np.zeros((self._num_kpts, 3))
        kpts_weight: np.ndarray = np.zeros(self._num_kpts)
        hsps: dict[str, np.ndarray] = {}
        for ii in range(num_rows):
            #  0.00000     0.00000    0.00000     0.03704           G
            tmp_row_lst: list[str] = linecache.getline(str(self.filename), row_idx + ii + 1).split()
            for jj in range(3):
                kpts[ii][jj] = float(tmp_row_lst[jj].strip())
            kpts_weight[ii] = float(tmp_row_lst[3].strip())

            if len(tmp_row_lst) == 5:
                hsps |= {
                    tmp_row_lst[4]: np.array(
                        [
                            float(tmp_row_lst[0]),
                            float(tmp_row_lst[1]),
                            float(tmp_row_lst[2]),
                        ]
                    )
                }
        return kpts, kpts_weight, hsps

    @property
    def spin(self) -> int:
        """The spin switches.

        Returns:
            int: Spin switches. 1 represents turn on spin, 2 represents turn down spin.
        """
        return self._spin

    @property
    def n_kpoints(self) -> int:
        """The number of k-points."""
        return self._num_kpts

    @property
    def n_bands(self) -> int:
        """The number of bands."""
        return self._num_bands

    @property
    def eigenvalues(self) -> np.ndarray:
        """The eigenvalues.

        Returns:
            np.ndarray: The first index represents spin, the second index
                represents kpoint, the third index represents band.
        """
        return self._eigenvalues

    @property
    def kpoints(self) -> np.ndarray:
        """The fractional coordinates of kpoints."""
        return self._kpts

    @property
    def kpoints_weight(self) -> np.ndarray:
        """The weight of kpoints."""
        return self._kpts_weight

    @property
    def hsps(self) -> dict[str, np.ndarray]:
        """The labels and fractional coordinates of high symmetry points as dict[str, np.ndarray].
        Empty dict when task is not line-mode kpath.
        """
        return self._hsps


class DosSpin(MSONable):
    """Extract information of DOS from DOS_SPIN file:
    - DOS.totalspin, DOS.totalspin_projected
    - DOS.spinup, DOS.spinup_projected
    - DOS.spindown, DOS.spindown_projected.
    """

    def __init__(self, filename: PathLike):
        self.filename: PathLike = filename
        self._labels, self._dos = self._parse()

    def _parse(self):
        """Parse the DOS_SPIN file to get name and values of partial DOS.

        Returns:
            labels (list[str]): The label of DOS, e.g. Total, Cr-3S, ...
            dos (NDArray): Value of density of state.
        """
        labels: list[str] = []
        labels = linecache.getline(str(self.filename), 1).split()[1:]
        dos_str: str = ""
        with zopen(self.filename, mode="rt", encoding="utf-8") as file:
            file.readline()
            dos_str = file.read()
        dos: np.ndarray = np.loadtxt(StringIO(dos_str))
        return labels, dos

    @property
    def labels(self) -> list[str]:
        """The name of the partial density of states."""
        return self._labels

    @property
    def dos(self) -> np.ndarray:
        """Value of density of state."""
        return self._dos

    def get_partial_dos(self, part: str) -> np.ndarray:
        """Get partial dos for give element or orbital.

        Args:
            part (str): The name of partial dos.
                e.g. 'Energy', 'Total', 'Cr-3S', 'Cr-3P',
                    'Cr-4S', 'Cr-3D', 'I-4D', 'I-5S', 'I-5P', 'Cr-3S', 'Cr-3Pz',
                    'Cr-3Px', 'Cr-3Py', 'Cr-4S', 'Cr-3Dz2','Cr-3Dxz', 'Cr-3Dyz',
                    'Cr-3D(x^2-y^2)', 'Cr-3Dxy', 'I-4Dz2', 'I-4Dxz', 'I-4Dyz',
                    'I-4D(x^2-y^2)', 'I-4Dxy', 'I-5S', 'I-5Pz', 'I-5Px', 'I-5Py'

        Returns:
            partial_dos: np.array
        """
        part_upper: str = part.upper()
        labels_upper: list[str] = [tmp_label.upper() for tmp_label in self._labels]
        idx_dos = labels_upper.index(part_upper)
        return self._dos[:, idx_dos]



"""
This module defines the VaspInputSet abstract base class and a concrete implementation for the parameters developed
and tested by the core team of pymatgen, including the Materials Virtual Lab, Materials Project and the MIT high
throughput project. The basic concept behind an input set is to specify a scheme to generate a consistent set of VASP
inputs from a structure without further user intervention. This ensures comparability across runs.

Read the following carefully before implementing new input sets:

1. 99% of what needs to be done can be done by specifying user_incar_settings to override some of the defaults of
   various input sets. Unless there is an extremely good reason to add a new set, **do not** add one. e.g. if you want
   to turn the Hubbard U off, just set "LDAU": False as a user_incar_setting.
2. All derivative input sets should inherit appropriate configurations (e.g., from MPRelaxSet), and more often than
   not, VaspInputSet should be the superclass. Superclass delegation should be used where possible. In particular,
   you are not supposed to implement your own as_dict or from_dict for derivative sets unless you know what you are
   doing. Improper overriding the as_dict and from_dict protocols is the major cause of implementation headaches. If
   you need an example, look at how the MPStaticSet is initialized.

The above are recommendations. The following are **UNBREAKABLE** rules:

1. All input sets must take in a structure, list of structures or None as the first argument. If None, the input set
   should perform a stateless initialization and before any output can be written, a structure must be set.
2. user_incar_settings, user_kpoints_settings and user_<whatever>_settings are ABSOLUTE. Any new sets you implement
   must obey this. If a user wants to override your settings, you assume he knows what he is doing. Do not
   magically override user supplied settings. You can issue a warning if you think the user is wrong.
3. All input sets must save all supplied args and kwargs as instance variables. e.g. self.arg = arg and
   self.kwargs = kwargs in the __init__. This ensures the as_dict and from_dict work correctly.
"""

from __future__ import annotations

import os
import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from pymatgen.io.lobster import Lobsterin
from pymatgen.io.vasp.sets import BadInputSetWarning, MPRelaxSet, VaspInputSet

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Literal

    UserPotcarFunctional = (
        Literal[
            "PBE",
            "PBE_52",
            "PBE_54",
            "PBE_64",
            "LDA",
            "LDA_52",
            "LDA_54",
            "PW91",
            "LDA_US",
            "PW91_US",
        ]
        | None
    )

MODULE_DIR = os.path.dirname(__file__)


@dataclass
class LobsterSet(VaspInputSet):
    """Input set to prepare VASP runs that can be digested by Lobster (See cohp.de).

    Args:
        structure (Structure): input structure.
        isym (int): ISYM entry for INCAR, only isym=-1 and isym=0 are allowed
        ismear (int): ISMEAR entry for INCAR, only ismear=-5 and ismear=0 are allowed
        reciprocal_density (int): density of k-mesh by reciprocal volume
        user_supplied_basis (dict): dict including basis functions for all elements in
            structure, e.g. {"Fe": "3d 3p 4s", "O": "2s 2p"}; if not supplied, a
            standard basis is used
        address_basis_file (str): address to a file similar to
            "BASIS_PBE_54_standard.yaml" in pymatgen.io.lobster.lobster_basis
        user_potcar_settings (dict): dict including potcar settings for all elements in
            structure, e.g. {"Fe": "Fe_pv", "O": "O"}; if not supplied, a standard basis
            is used.
        **kwargs: Other kwargs supported by VaspInputSet.
    """

    isym: int = 0
    ismear: int = -5
    reciprocal_density: int | None = None
    address_basis_file: str | None = None
    user_supplied_basis: dict | None = None

    # Latest POTCARs are preferred
    # Choose PBE_54 unless the user specifies a different potcar_functional
    user_potcar_functional: UserPotcarFunctional = "PBE_54"

    CONFIG = MPRelaxSet.CONFIG
    _valid_potcars: Sequence[str] | None = ("PBE_52", "PBE_54", "PBE_64")

    def __post_init__(self) -> None:
        super().__post_init__()
        warnings.warn(
            "Make sure that all parameters are okay! This is a brand new implementation.",
            stacklevel=2,
        )

        if self.user_potcar_functional in ["PBE_52", "PBE_64"]:
            warnings.warn(
                f"Using {self.user_potcar_functional} POTCARs with basis functions generated for PBE_54 POTCARs. "
                "Basis functions for elements with obsoleted, updated or newly added POTCARs in "
                f"{self.user_potcar_functional} will not be available and may cause errors or inaccuracies.",
                BadInputSetWarning,
                stacklevel=2,
            )
        if self.isym not in {-1, 0}:
            raise ValueError("Lobster cannot digest WAVEFUNCTIONS with symmetry. isym must be -1 or 0")
        if self.ismear not in {-5, 0}:
            raise ValueError("Lobster usually works with ismear=-5 or ismear=0")

        self._config_dict["POTCAR"]["W"] = "W_sv"

    @property
    def kpoints_updates(self) -> dict[str, int]:
        """Updates to the kpoints configuration for this calculation type."""
        # Test if this is okay
        return {"reciprocal_density": self.reciprocal_density or 310}

    @property
    def incar_updates(self) -> dict[str, Any]:
        """Updates to the INCAR config for this calculation type."""

        potcar_symbols = self.potcar_symbols

        # Predefined basis! Check if the basis is okay! (charge spilling and bandoverlaps!)
        if self.user_supplied_basis is None and self.address_basis_file is None:
            basis = Lobsterin.get_basis(structure=self.structure, potcar_symbols=potcar_symbols)  # type: ignore[arg-type]
        elif self.address_basis_file is not None:
            basis = Lobsterin.get_basis(
                structure=self.structure,  # type: ignore[arg-type]
                potcar_symbols=potcar_symbols,
                address_basis_file=self.address_basis_file,
            )
        elif self.user_supplied_basis is not None:
            # Test if all elements from structure are in user_supplied_basis
            for atom_type in self.structure.symbol_set:  # type: ignore[union-attr]
                if atom_type not in self.user_supplied_basis:
                    raise ValueError(f"There are no basis functions for the atom type {atom_type}")
            basis = [f"{key} {value}" for key, value in self.user_supplied_basis.items()]
        else:
            basis = None

        lobsterin = Lobsterin(settingsdict={"basisfunctions": basis})
        nbands = lobsterin._get_nbands(structure=self.structure)  # type: ignore[arg-type]

        return {
            "EDIFF": 1e-6,
            "NSW": 0,
            "LWAVE": True,
            "ISYM": self.isym,
            "NBANDS": nbands,
            "IBRION": -1,
            "ISMEAR": self.ismear,
            "LORBIT": 11,
            "ICHARG": 0,
            "ALGO": "Normal",
        }



"""
This module implements input and output for Fiesta (http://perso.neel.cnrs.fr/xavier.blase/fiesta/index.html).

- Nwchem2Fiesta: create the input files needed for a Fiesta run
- FiestaRun: run gw_fiesta and bse_fiesta
- Localized BasisSetReader
"""

from __future__ import annotations

import os
import re
import shutil
import subprocess
from string import Template
from typing import TYPE_CHECKING

from monty.io import zopen
from monty.json import MSONable

from pymatgen.core.structure import Molecule

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Self


__author__ = "ndardenne"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__email__ = "n.dardenne@uclouvain.be"
__date__ = "24/5/15"


class Nwchem2Fiesta(MSONable):
    """To run NWCHEM2FIESTA inside python:

    If nwchem.nw is the input, nwchem.out the output, and structure.movecs the
    "movecs" file, the syntax to run NWCHEM2FIESTA is: NWCHEM2FIESTA
    nwchem.nw nwchem.nwout structure.movecs > log_n2f
    """

    def __init__(self, folder, filename="nwchem", log_file="log_n2f"):
        """
        Args:
            folder: where are stored the nwchem
            filename: name of nwchem files read by NWCHEM2FIESTA (filename.nw, filename.nwout and filename.movecs)
            logfile: logfile of NWCHEM2FIESTA.
        """
        self.folder = folder
        self.filename = filename
        self.log_file = log_file

        self._NWCHEM2FIESTA_cmd = "NWCHEM2FIESTA"
        self._nwcheminput_fn = f"{filename}.nw"
        self._nwchemoutput_fn = f"{filename}.nwout"
        self._nwchemmovecs_fn = f"{filename}.movecs"

    def run(self):
        """Perform actual NWCHEM2FIESTA run."""
        init_folder = os.getcwd()
        os.chdir(self.folder)

        with zopen(self.log_file, mode="wt", encoding="utf-8") as fout:
            subprocess.call(
                [
                    self._NWCHEM2FIESTA_cmd,
                    self._nwcheminput_fn,
                    self._nwchemoutput_fn,
                    self._nwchemmovecs_fn,
                ],
                stdout=fout,
            )

        os.chdir(init_folder)

    def as_dict(self):
        """MSONable dict."""
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "filename": self.filename,
            "folder": self.folder,
        }

    @classmethod
    def from_dict(cls, dct: dict) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            Nwchem2Fiesta
        """
        return cls(folder=dct["folder"], filename=dct["filename"])


class FiestaRun(MSONable):
    """
    To run FIESTA inside python:
        if grid is [x,x] then bse runs
        if grid is [x,x,y] the fiesta(gw) runs
        otherwise it breaks.
    """

    def __init__(
        self,
        folder: str | None = None,
        grid: tuple[int, int, int] = (2, 2, 2),
        log_file: str = "log",
    ) -> None:
        """
        Args:
            folder: Folder to look for runs.
            grid:
            log_file: logfile of Fiesta.
        """
        self.folder = folder or os.getcwd()
        self.log_file = log_file
        self.grid = grid

    def run(self):
        """Perform FIESTA (gw) run."""
        if len(self.grid) == 3:
            self.mpi_procs = self.grid[0] * self.grid[1] * self.grid[2]
            self._gw_run()
        elif len(self.grid) == 2:
            self.mpi_procs = self.grid[0] * self.grid[1]
            self.bse_run()
        else:
            raise ValueError("Wrong grid size: must be [nrow, ncolumn, nslice] for gw of [nrow, nslice] for bse")

    def _gw_run(self):
        """Perform FIESTA (gw) run."""
        init_folder = os.getcwd()
        if self.folder != init_folder:
            os.chdir(self.folder)

        with zopen(self.log_file, mode="wt", encoding="utf-8") as fout:
            subprocess.call(
                [
                    "mpirun",
                    "-n",
                    str(self.mpi_procs),
                    "fiesta",
                    str(self.grid[0]),
                    str(self.grid[1]),
                    str(self.grid[2]),
                ],
                stdout=fout,
            )

        if self.folder != init_folder:
            os.chdir(init_folder)

    def bse_run(self):
        """Perform BSE run."""
        init_folder = os.getcwd()
        if self.folder != init_folder:
            os.chdir(self.folder)

        with zopen(self.log_file, mode="wt", encoding="utf-8") as fout:
            subprocess.call(
                [
                    "mpirun",
                    "-n",
                    str(self.mpi_procs),
                    "bse",
                    str(self.grid[0]),
                    str(self.grid[1]),
                ],
                stdout=fout,
            )

        if self.folder != init_folder:
            os.chdir(init_folder)

    def as_dict(self):
        """MSONable dict."""
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "log_file": self.log_file,
            "grid": self.grid,
            "folder": self.folder,
        }

    @classmethod
    def from_dict(cls, dct: dict) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            FiestaRun
        """
        return cls(folder=dct["folder"], grid=dct["grid"], log_file=dct["log_file"])


class BasisSetReader:
    """
    A basis set reader.
    Basis set are stored in data as a dict:
    :key l_zeta_ng for each nl orbitals which contain list of tuple (alpha, coef) for each of the ng gaussians
    in l_zeta orbital.
    """

    def __init__(self, filename):
        """
        Args:
            filename: Filename to read.
        """
        self.filename = filename

        with zopen(filename, mode="rt", encoding="utf-8") as file:
            basis_set = file.read()

        self.data = self._parse_file(basis_set)
        # compute the number of nlm orbitals per atom
        self.data.update(n_nlmo=self.set_n_nlmo())

    @staticmethod
    def _parse_file(lines):
        lmax_nnlo_patt = re.compile(r"\s* (\d+) \s+ (\d+) \s+ \# .* ", re.VERBOSE)

        nl_orbital_patt = re.compile(r"\s* (\d+) \s+ (\d+) \s+ (\d+) \s+ \# .* ", re.VERBOSE)

        coef_alpha_patt = re.compile(r"\s* ([-\d.\D]+) \s+ ([-\d.\D]+) \s* ", re.VERBOSE)

        preamble = []
        basis_set = {}
        parse_preamble = parse_lmax_nnlo = False
        parse_nl_orbital = False
        nnlo = lmax = None
        l_angular = zeta = ng = None

        for line in lines.split("\n"):
            if parse_nl_orbital:
                match_orb = nl_orbital_patt.search(line)
                match_alpha = coef_alpha_patt.search(line)
                if match_orb:
                    l_angular = match_orb[1]
                    zeta = match_orb[2]
                    ng = match_orb[3]
                    basis_set[f"{l_angular}_{zeta}_{ng}"] = []
                elif match_alpha:
                    alpha = match_alpha[1]
                    coef = match_alpha[2]
                    basis_set[f"{l_angular}_{zeta}_{ng}"].append((alpha, coef))
            elif parse_lmax_nnlo:
                if match_orb := lmax_nnlo_patt.search(line):
                    lmax = match_orb[1]
                    nnlo = match_orb[2]
                    parse_lmax_nnlo = False
                    parse_nl_orbital = True
            elif parse_preamble:
                preamble.append(line.strip())

            if line.find("</preamble>") != -1:
                parse_preamble = False
                parse_lmax_nnlo = True
            elif line.find("<preamble>") != -1:
                parse_preamble = True

        basis_set.update(lmax=lmax, n_nlo=nnlo, preamble=preamble)
        return basis_set

    def set_n_nlmo(self):
        """The number of nlm orbitals for the basis set."""
        n_nlm_orbs = 0

        data_tmp = self.data
        data_tmp.pop("lmax")
        data_tmp.pop("n_nlo")
        data_tmp.pop("preamble")

        for l_zeta_ng in data_tmp:
            n_l = l_zeta_ng.split("_")[0]
            n_nlm_orbs += 2 * int(n_l) + 1

        return str(n_nlm_orbs)

    def infos_on_basis_set(self):
        return (
            f"=========================================\n"
            f"Reading basis set:\n"
            f"\n"
            f"Basis set for {self.filename} atom \n"
            f"Maximum angular momentum = {self.data['lmax']}\n"
            f"Number of atomics orbitals = {self.data['n_nlo']}\n"
            f"Number of nlm orbitals = {self.data['n_nlmo']}\n"
            f"========================================="
        )


class FiestaInput(MSONable):
    """Input File for Fiesta called "cell.in" by default (mandatory in Fiesta for now)."""

    def __init__(
        self,
        mol,
        correlation_grid: dict[str, str] | None = None,
        exc_dft_option: dict[str, str] | None = None,
        cohsex_options: dict[str, str] | None = None,
        gw_options: dict[str, str] | None = None,
        bse_tddft_options: dict[str, str] | None = None,
    ):
        """
        Args:
            mol: pymatgen mol
            correlation_grid: dict
            Exc_DFT_option: dict
            COHSEX_options: dict
            GW_options: dict
            BSE_TDDFT_options: dict.
        """
        self._mol = mol
        self.correlation_grid = correlation_grid or {"dE_grid": "0.500", "n_grid": "14"}
        self.Exc_DFT_option = exc_dft_option or {"rdVxcpsi": "1"}
        self.cohsex_options = cohsex_options or {
            "eigMethod": "C",
            "mix_cohsex": "0.500",
            "nc_cohsex": "0",
            "nit_cohsex": "0",
            "nv_cohsex": "0",
            "resMethod": "V",
            "scf_cohsex_wf": "0",
        }
        self.GW_options = gw_options or {
            "nc_corr": "10",
            "nit_gw": "3",
            "nv_corr": "10",
        }
        self.bse_tddft_options = bse_tddft_options or {
            "do_bse": "1",
            "do_tddft": "0",
            "nc_bse": "382",
            "nit_bse": "50",
            "npsi_bse": "1",
            "nv_bse": "21",
        }

    def set_auxiliary_basis_set(self, folder, auxiliary_folder, auxiliary_basis_set_type="aug_cc_pvtz"):
        """
        copy in the desired folder the needed auxiliary basis set "X2.ion" where X is a specie.

        Args:
            auxiliary_folder: folder where the auxiliary basis sets are stored
            auxiliary_basis_set_type: type of basis set (string to be found in the extension of the file name; must
                be in lower case). ex: C2.ion_aug_cc_pvtz_RI_Weigend find "aug_cc_pvtz".
        """
        list_files = os.listdir(auxiliary_folder)

        for specie in self._mol.symbol_set:
            for file in list_files:
                if file.upper().find(f"{specie.upper()}2") != -1 and file.lower().find(auxiliary_basis_set_type) != -1:
                    shutil.copyfile(f"{auxiliary_folder}/{file}", f"{folder}/{specie}2.ion")

    def set_gw_options(self, nv_band=10, nc_band=10, n_iteration=5, n_grid=6, dE_grid=0.5):
        """Set parameters in cell.in for a GW computation.

        Args:
            nv__band: number of valence bands to correct with GW
            nc_band: number of conduction bands to correct with GW
            n_iteration: number of iteration
            n_grid and dE_grid: number of points and spacing in eV for correlation grid.
        """
        self.GW_options.update(nv_corr=nv_band, nc_corr=nc_band, nit_gw=n_iteration)
        self.correlation_grid.update(dE_grid=dE_grid, n_grid=n_grid)

    @staticmethod
    def make_full_bse_densities_folder(folder):
        """Mkdir "FULL_BSE_Densities" folder (needed for bse run) in the desired folder."""
        if os.path.isfile(f"{folder}/FULL_BSE_Densities"):
            return "FULL_BSE_Densities folder already exists"

        os.makedirs(f"{folder}/FULL_BSE_Densities")
        return "makedirs FULL_BSE_Densities folder"

    def set_bse_options(self, n_excitations=10, nit_bse=200):
        """Set parameters in cell.in for a BSE computation.

        Args:
            nv_bse: number of valence bands
            nc_bse: number of conduction bands
            n_excitations: number of excitations
            nit_bse: number of iterations.
        """
        self.bse_tddft_options.update(npsi_bse=n_excitations, nit_bse=nit_bse)

    def dump_bse_data_in_gw_run(self, BSE_dump=True):
        """Set the "do_bse" variable to 1 or 0 in cell.in.

        Args:
            BSE_dump (bool): Defaults to True.
        """
        self.bse_tddft_options.update(do_bse=int(BSE_dump), do_tddft=0)

    def dump_tddft_data_in_gw_run(self, tddft_dump: bool = True):
        """Set the do_tddft variable to 1 or 0 in cell.in.

        Args:
            tddft_dump (bool): Defaults to True.
        """
        self.bse_tddft_options.update(do_bse="0", do_tddft="1" if tddft_dump else "0")

    @property
    def infos_on_system(self):
        """Infos on initial parameters as in the log file of Fiesta."""
        lst = [
            "=========================================",
            "Reading infos on system:",
            "",
            f" Number of atoms = {self._mol.composition.num_atoms} ; number of species = {len(self._mol.symbol_set)}",
            f" Number of valence bands = {int(self._mol.nelectrons / 2)}",
            f" Sigma grid specs: n_grid = {self.correlation_grid['n_grid']} ;  "
            f"dE_grid = {self.correlation_grid['dE_grid']} (eV)",
        ]
        if int(self.Exc_DFT_option["rdVxcpsi"]) == 1:
            lst.append(" Exchange and correlation energy read from Vxcpsi.mat")
        elif int(self.Exc_DFT_option["rdVxcpsi"]) == 0:
            lst.append(" Exchange and correlation energy re-computed")

        if self.cohsex_options["eigMethod"] == "C":
            lst += [
                f" Correcting  {self.cohsex_options['nv_cohsex']} valence bands and  "
                f"{self.cohsex_options['nc_cohsex']} conduction bands at COHSEX level",
                f" Performing   {self.cohsex_options['nit_cohsex']} diagonal COHSEX iterations",
            ]
        elif self.cohsex_options["eigMethod"] == "HF":
            lst += [
                f" Correcting  {self.cohsex_options['nv_cohsex']} valence bands and  "
                f"{self.cohsex_options['nc_cohsex']} conduction bands at HF level",
                f" Performing   {self.cohsex_options['nit_cohsex']} diagonal HF iterations",
            ]

        lst += [
            f" Using resolution of identity : {self.cohsex_options['resMethod']}",
            f" Correcting  {self.GW_options['nv_corr']} valence bands and "
            f"{self.GW_options['nc_corr']} conduction bands at GW level",
            f" Performing   {self.GW_options['nit_gw']} GW iterations",
        ]

        if int(self.bse_tddft_options["do_bse"]) == 1:
            lst.append(" Dumping data for BSE treatment")

        if int(self.bse_tddft_options["do_tddft"]) == 1:
            lst.append(" Dumping data for TD-DFT treatment")
        lst.extend(("", " Atoms in cell cartesian A:"))
        symbols = list(self._mol.symbol_set)

        for site in self._mol:
            lst.append(f" {site.x} {site.y} {site.z} {int(symbols.index(site.specie.symbol)) + 1}")

        lst.append("=========================================")

        return str(lst)

    @property
    def molecule(self):
        """Molecule associated with this FiestaInput."""
        return self._mol

    def __str__(self):
        symbols = list(self._mol.symbol_set)

        geometry = [f" {site.x} {site.y} {site.z} {symbols.index(site.specie.symbol) + 1}" for site in self._mol]

        t = Template(
            """# number of atoms and species
   $nat    $nsp
# number of valence bands
    $nvbands
# number of points and spacing in eV for correlation grid
    $n_grid    $dE_grid
# relire=1 ou recalculer=0 Exc DFT
    $rdVxcpsi
# number of COHSEX corrected occp and unoccp bands: C=COHSEX  H=HF
    $nv_cohsex    $nc_cohsex   $eigMethod
# number of COHSEX iter, scf on wfns, mixing coeff; V=RI-V  I=RI-D
    $nit_cohsex   $resMethod       $scf_cohsex_wf       $mix_cohsex
# number of GW corrected occp and unoccp bands
   $nv_corr   $nc_corr
# number of GW iterations
    $nit_gw
# dumping for BSE and TDDFT
    $do_bse    $do_tddft
# number of occp. and virtual bands of BSE: nocore and up to 40 eVs
    $nv_bse   $nc_bse
# number of excitations needed and number of iterations
    $npsi_bse   $nit_bse
# list of symbols in order
$symbols
# scaling factor
    1.000
# atoms x,y,z cartesian .. will be multiplied by scale
$geometry
            """
        )

        return t.substitute(
            nat=int(self._mol.composition.num_atoms),
            nsp=len(self._mol.symbol_set),
            nvbands=int(self._mol.nelectrons / 2),
            n_grid=self.correlation_grid["n_grid"],
            dE_grid=self.correlation_grid["dE_grid"],
            rdVxcpsi=self.Exc_DFT_option["rdVxcpsi"],
            nv_cohsex=self.cohsex_options["nv_cohsex"],
            nc_cohsex=self.cohsex_options["nc_cohsex"],
            eigMethod=self.cohsex_options["eigMethod"],
            nit_cohsex=self.cohsex_options["nit_cohsex"],
            resMethod=self.cohsex_options["resMethod"],
            scf_cohsex_wf=self.cohsex_options["scf_cohsex_wf"],
            mix_cohsex=self.cohsex_options["mix_cohsex"],
            nv_corr=self.GW_options["nv_corr"],
            nc_corr=self.GW_options["nc_corr"],
            nit_gw=self.GW_options["nit_gw"],
            do_bse=self.bse_tddft_options["do_bse"],
            do_tddft=self.bse_tddft_options["do_tddft"],
            nv_bse=self.bse_tddft_options["nv_bse"],
            nc_bse=self.bse_tddft_options["nc_bse"],
            npsi_bse=self.bse_tddft_options["npsi_bse"],
            nit_bse=self.bse_tddft_options["nit_bse"],
            symbols="\n".join(symbols),
            geometry="\n".join(geometry),
        )

    def write_file(self, filename: str | Path) -> None:
        """Write FiestaInput to a file.

        Args:
            filename: Filename.
        """
        with zopen(filename, mode="wt", encoding="utf-8") as file:
            file.write(str(self))  # type:ignore[arg-type]

    def as_dict(self):
        """MSONable dict."""
        return {
            "mol": self._mol.as_dict(),
            "correlation_grid": self.correlation_grid,
            "Exc_DFT_option": self.Exc_DFT_option,
            "COHSEX_options": self.cohsex_options,
            "GW_options": self.GW_options,
            "BSE_TDDFT_options": self.bse_tddft_options,
        }

    @classmethod
    def from_dict(cls, dct: dict) -> Self:
        """
        Args:
            dct (dict): Dict representation.

        Returns:
            FiestaInput
        """
        return cls(
            mol=Molecule.from_dict(dct["mol"]),
            correlation_grid=dct["correlation_grid"],
            exc_dft_option=dct["Exc_DFT_option"],
            cohsex_options=dct["geometry_options"],
            gw_options=dct["symmetry_options"],
            bse_tddft_options=dct["memory_options"],
        )

    @classmethod
    def from_str(cls, string_input: str) -> FiestaInput:
        """
        Read an FiestaInput from a string. Currently tested to work with
        files generated from this class itself.

        Args:
            string_input: string_input to parse.

        Returns:
            FiestaInput object
        """
        correlation_grid = {}
        Exc_DFT_option = {}
        COHSEX_options = {}
        GW_options = {}
        BSE_TDDFT_options = {}

        lines = string_input.strip().split("\n")

        # number of atoms and species
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        nat = tokens[0]
        nsp = tokens[1]
        # number of valence bands
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()

        # correlation_grid
        # number of points and spacing in eV for correlation grid
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        correlation_grid["n_grid"] = tokens[0]
        correlation_grid["dE_grid"] = tokens[1]

        # Exc DFT
        # relire=1 ou recalculer=0 Exc DFT
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        Exc_DFT_option["rdVxcpsi"] = tokens[0]

        # COHSEX
        # number of COHSEX corrected occp and unoccp bands: C=COHSEX  H=HF
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        COHSEX_options["nv_cohsex"] = tokens[0]
        COHSEX_options["nc_cohsex"] = tokens[1]
        COHSEX_options["eigMethod"] = tokens[2]
        # number of COHSEX iter, scf on wfns, mixing coeff; V=RI-V  I=RI-D
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        COHSEX_options["nit_cohsex"] = tokens[0]
        COHSEX_options["resMethod"] = tokens[1]
        COHSEX_options["scf_cohsex_wf"] = tokens[2]
        COHSEX_options["mix_cohsex"] = tokens[3]

        # GW
        # number of GW corrected occp and unoccp bands
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        GW_options["nv_corr"] = tokens[0]
        GW_options["nc_corr"] = tokens[1]
        # number of GW iterations
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        GW_options["nit_gw"] = tokens[0]

        # BSE
        # dumping for BSE and TDDFT
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        BSE_TDDFT_options["do_bse"] = tokens[0]
        BSE_TDDFT_options["do_tddft"] = tokens[1]
        # number of occp. and virtual bands of BSE: nocore and up to 40 eVs
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        BSE_TDDFT_options["nv_bse"] = tokens[0]
        BSE_TDDFT_options["nc_bse"] = tokens[1]
        # number of excitations needed and number of iterations
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        BSE_TDDFT_options["npsi_bse"] = tokens[0]
        BSE_TDDFT_options["nit_bse"] = tokens[1]

        # Molecule
        # list of symbols in order
        lines.pop(0)
        atname = []
        i = int(nsp)
        while i != 0:
            line = lines.pop(0).strip()
            tokens = line.split()
            atname.append(tokens[0])
            i -= 1

        # scaling factor
        lines.pop(0)
        line = lines.pop(0).strip()
        tokens = line.split()
        # atoms x,y,z cartesian .. will be multiplied by scale
        lines.pop(0)
        # Parse geometry
        species = []
        coords = []
        i = int(nat)
        while i != 0:
            line = lines.pop(0).strip()
            tokens = line.split()
            coords.append([float(j) for j in tokens[:3]])
            species.append(atname[int(tokens[3]) - 1])
            i -= 1

        mol = Molecule(species, coords)

        return FiestaInput(
            mol=mol,
            correlation_grid=correlation_grid,
            exc_dft_option=Exc_DFT_option,
            cohsex_options=COHSEX_options,
            gw_options=GW_options,
            bse_tddft_options=BSE_TDDFT_options,
        )

    @classmethod
    def from_file(cls, filename: str | Path) -> FiestaInput:
        """
        Read an Fiesta input from a file. Currently tested to work with
        files generated from this class itself.

        Args:
            filename: Filename to parse.

        Returns:
            FiestaInput object
        """
        with zopen(filename, mode="rt", encoding="utf-8") as file:
            return cls.from_str(file.read())  # type:ignore[arg-type]


class FiestaOutput:
    """
    A Fiesta output file parser.

    All energies are in eV.
    """

    def __init__(self, filename):
        """
        Args:
            filename: Filename to read.
        """
        self.filename = filename

        with zopen(filename, mode="rt", encoding="utf-8") as file:
            data = file.read()

        chunks = re.split(r"GW Driver iteration", data)

        # preamble: everything before the first GW Driver iteration
        chunks.pop(0)

        # self.job_info = self._parse_preamble(preamble)
        self.data = [self._parse_job(c) for c in chunks]

    @staticmethod
    def _parse_job(output):
        GW_BANDS_results_patt = re.compile(
            r"^<it.*  \| \s+ (\D+\d*) \s+ \| \s+ ([-\d.]+) \s+ ([-\d.]+) \s+ ([-\d.]+) \s+ \| "
            r" \s+ ([-\d.]+) \s+ ([-\d.]+) \s+ ([-\d.]+) \s+ \|"
            r" \s+ ([-\d.]+) \s+ ([-\d.]+) \s+ ",
            re.VERBOSE,
        )

        GW_GAPS_results_patt = re.compile(
            r"^<it.*  \| \s+ Egap_KS \s+ = \s+ ([-\d.]+) \s+ \| \s+ Egap_QP \s+ = \s+ ([-\d.]+) \s+ \| "
            r" \s+ Egap_QP \s+ = \s+ ([-\d.]+) \s+ \|",
            re.VERBOSE,
        )

        end_patt = re.compile(r"\s*program returned normally\s*")

        total_time_patt = re.compile(r"\s*total \s+ time: \s+  ([\d.]+) .*", re.VERBOSE)

        GW_results = {}
        parse_gw_results = parse_total_time = False

        for line in output.split("\n"):
            if parse_total_time:
                if match := end_patt.search(line):
                    GW_results.update(end_normally=True)

                if match := total_time_patt.search(line):
                    GW_results.update(total_time=match[1])

            if parse_gw_results:
                if line.find("Dumping eigen energies") != -1:
                    parse_total_time = True
                    parse_gw_results = False
                    continue

                if match := GW_BANDS_results_patt.search(line):
                    dct = {}
                    dct.update(
                        band=match[1].strip(),
                        eKS=match[2],
                        eXX=match[3],
                        eQP_old=match[4],
                        z=match[5],
                        sigma_c_Linear=match[6],
                        eQP_Linear=match[7],
                        sigma_c_SCF=match[8],
                        eQP_SCF=match[9],
                    )
                    GW_results[match[1].strip()] = dct

                if n := GW_GAPS_results_patt.search(line):
                    dct = {}
                    dct.update(
                        Egap_KS=n[1],
                        Egap_QP_Linear=n[2],
                        Egap_QP_SCF=n[3],
                    )
                    GW_results["Gaps"] = dct

            if line.find("GW Results") != -1:
                parse_gw_results = True

        return GW_results


class BSEOutput:
    """
    A bse output file parser. The start...

    All energies are in eV.
    """

    def __init__(self, filename):
        """
        Args:
            filename: Filename to read.
        """
        self.filename = filename

        with zopen(filename, mode="rt", encoding="utf-8") as file:
            log_bse = file.read()

        # self.job_info = self._parse_preamble(preamble)
        self.exiton = self._parse_job(log_bse)

    @staticmethod
    def _parse_job(output):
        BSE_exitons_patt = re.compile(
            r"^exiton \s+ (\d+)  : \s+  ([\d.]+) \( \s+ ([-\d.]+) \) \s+ \| .*  ",
            re.VERBOSE,
        )

        end_patt = re.compile(r"\s*program returned normally\s*")

        total_time_patt = re.compile(r"\s*total \s+ time: \s+  ([\d.]+) .*", re.VERBOSE)

        BSE_results = {}
        parse_BSE_results = parse_total_time = False

        for line in output.split("\n"):
            if parse_total_time:
                if match := end_patt.search(line):
                    BSE_results.update(end_normally=True)

                if match := total_time_patt.search(line):
                    BSE_results.update(total_time=match[1])

            if parse_BSE_results:
                if line.find("FULL BSE main valence -> conduction transitions weight:") != -1:
                    parse_total_time = True
                    parse_BSE_results = False
                    continue

                if match := BSE_exitons_patt.search(line):
                    dct = {}
                    dct.update(bse_eig=match[2], osc_strength=match[3])
                    BSE_results[str(match[1].strip())] = dct

            if line.find("FULL BSE eig.(eV), osc. strength and dipoles:") != -1:
                parse_BSE_results = True

        return BSE_results



"""Module for reading Lobster output files. For more information on LOBSTER see www.cohp.de.

If you use this module, please cite:

J. George, G. Petretto, A. Naik, M. Esters, A. J. Jackson, R. Nelson, R. Dronskowski, G.-M. Rignanese, G. Hautier,
"Automated Bonding Analysis with Crystal Orbital Hamilton Populations", ChemPlusChem 2022, e202200123,
DOI: 10.1002/cplu.202200123.
"""

from __future__ import annotations

from pymatgen.io.lobster.future.outputs.bands import BandOverlaps, Fatband, Fatbands
from pymatgen.io.lobster.future.outputs.coxxcar import COBICAR, COBICAR_LCFO, COHPCAR, COHPCAR_LCFO, COOPCAR, COXXCAR
from pymatgen.io.lobster.future.outputs.doscar import DOSCAR, DOSCAR_LCFO
from pymatgen.io.lobster.future.outputs.icoxxlist import (
    ICOBILIST,
    ICOBILIST_LCFO,
    ICOHPLIST,
    ICOHPLIST_LCFO,
    ICOOPLIST,
    ICOXXLIST,
    NcICOBILIST,
)
from pymatgen.io.lobster.future.outputs.lobsterout import LobsterOut
from pymatgen.io.lobster.future.outputs.misc import (
    BWDF,
    BWDFCOHP,
    POLARIZATION,
    LobsterMatrices,
    MadelungEnergies,
    SitePotentials,
    Wavefunction,
)
from pymatgen.io.lobster.future.outputs.populations import CHARGE, CHARGE_LCFO, GROSSPOP, GROSSPOP_LCFO
from pymatgen.util.due import Doi, due  # type: ignore[import]

__author__ = "Tom Demeyere"
__copyright__ = "Copyright 2025, The Materials Project"
__version__ = "0.3"
__maintainer__ = "Tom Demeyere"
__email__ = "tom.demeyere@bam.de"
__date__ = "Sep. 30, 2025"

due.cite(
    Doi("10.1002/cplu.202200123"),
    description="Automated Bonding Analysis with Crystal Orbital Hamilton Populations",
)

__all__ = [
    "BWDF",
    "BWDFCOHP",
    "CHARGE",
    "CHARGE_LCFO",
    "COBICAR",
    "COBICAR_LCFO",
    "COHPCAR",
    "COHPCAR_LCFO",
    "COOPCAR",
    "COXXCAR",
    "DOSCAR",
    "DOSCAR_LCFO",
    "GROSSPOP",
    "GROSSPOP_LCFO",
    "ICOBILIST",
    "ICOBILIST_LCFO",
    "ICOHPLIST",
    "ICOHPLIST_LCFO",
    "ICOOPLIST",
    "ICOXXLIST",
    "POLARIZATION",
    "BandOverlaps",
    "Fatband",
    "Fatbands",
    "LobsterMatrices",
    "LobsterOut",
    "MadelungEnergies",
    "NcICOBILIST",
    "SitePotentials",
    "Wavefunction",
]



"""Support for reading XCrysDen files."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pymatgen.core import Element, IStructure, Structure

if TYPE_CHECKING:
    from typing import Self

    from pymatgen.core.structure import IStructure

__author__ = "Matteo Giantomassi"
__copyright__ = "Copyright 2013, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Matteo Giantomassi"


class XSF:
    """Parse XCrysden files."""

    def __init__(self, structure: Structure | IStructure):
        """
        Args:
            structure (Structure): Structure object.
        """
        self.structure = structure

    def to_str(self, atom_symbol: bool = True) -> str:
        """Get a string with the structure in XSF format
        See http://www.xcrysden.org/doc/XSF.html.

        Args:
            atom_symbol (bool): Uses atom symbol instead of atomic number. Defaults to True.
        """
        lines: list[str] = []

        lines.extend(("CRYSTAL", "# Primitive lattice vectors in Angstrom", "PRIMVEC"))
        cell = self.structure.lattice.matrix
        lines.extend(f" {cell[i][0]:.14f} {cell[i][1]:.14f} {cell[i][2]:.14f}" for i in range(3))

        cart_coords = self.structure.cart_coords
        lines.extend(
            (
                "# Cartesian coordinates in Angstrom.",
                "PRIMCOORD",
                f" {len(cart_coords)} 1",
            )
        )

        for site, coord in zip(self.structure, cart_coords, strict=True):
            sp = site.specie.symbol if atom_symbol else f"{site.specie.Z}"
            x, y, z = coord
            lines.append(f"{sp} {x:20.14f} {y:20.14f} {z:20.14f}")
            if "vect" in site.properties:
                vx, vy, vz = site.properties["vect"]
                lines[-1] += f" {vx:20.14f} {vy:20.14f} {vz:20.14f}"

        return "\n".join(lines)

    @classmethod
    def from_str(cls, input_string: str, cls_=None) -> Self:
        """
        Initialize a `Structure` object from a string with data in XSF format.

        Args:
            input_string: String with the structure in XSF format.
                See http://www.xcrysden.org/doc/XSF.html
            cls_: Structure class to be created. default: pymatgen structure

        Example file:
            CRYSTAL                                        see (1)
            these are primitive lattice vectors (in Angstroms)
            PRIMVEC
            0.0000000    2.7100000    2.7100000         see (2)
            2.7100000    0.0000000    2.7100000
            2.7100000    2.7100000    0.0000000

            these are conventional lattice vectors (in Angstroms)
            CONVVEC
            5.4200000    0.0000000    0.0000000         see (3)
            0.0000000    5.4200000    0.0000000
            0.0000000    0.0000000    5.4200000

            these are atomic coordinates in a primitive unit cell  (in Angstroms)
            PRIMCOORD
            2 1                                            see (4)
            16      0.0000000     0.0000000     0.0000000  see (5)
            30      1.3550000    -1.3550000    -1.3550000
        """
        lattice: list[list[float]] = []
        coords: list[list[float]] = []
        species: list[int] = []
        lines = input_string.splitlines()

        for idx, line in enumerate(lines, start=1):
            if "PRIMVEC" in line:
                lattice.extend([float(c) for c in lines[j].split()] for j in range(idx, idx + 3))

            if "PRIMCOORD" in line:
                n_sites = int(lines[idx].split()[0])

                for j in range(idx + 1, idx + 1 + n_sites):
                    tokens = lines[j].split()
                    species.append(Element(tokens[0]).Z if tokens[0].isalpha() else int(tokens[0]))
                    coords.append([float(j) for j in tokens[1:4]])
                break
        else:
            raise ValueError("Invalid XSF data")

        if cls_ is None:
            cls_ = Structure

        return cls(cls_(lattice, species, coords, coords_are_cartesian=True))



"""Pymatgen package configuration."""

from __future__ import annotations

import platform
import sys

import numpy as np
from setuptools import Extension, setup

is_win_64 = sys.platform.startswith("win") and platform.machine().endswith("64")
extra_link_args = ["-Wl,--allow-multiple-definition"] if is_win_64 else []

setup(
    ext_modules=[
        Extension(
            "pymatgen.util.coord_cython",
            ["src/pymatgen/util/coord_cython.pyx"],
            extra_link_args=extra_link_args,
        ),
        Extension(
            "pymatgen.optimization.neighbors",
            ["src/pymatgen/optimization/neighbors.pyx"],
            extra_link_args=extra_link_args,
        ),
    ],
    include_dirs=[np.get_include()],
)



from __future__ import annotations

import re
from collections import defaultdict
from itertools import islice
from typing import TYPE_CHECKING

import numpy as np

from pymatgen.core.periodic_table import Element
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.dos import Dos
from pymatgen.io.lobster.future.core import LobsterFile
from pymatgen.io.lobster.future.versioning import version_processor

if TYPE_CHECKING:
    from typing import ClassVar

    from numpy import floating
    from numpy.typing import NDArray

    from pymatgen.util.typing import PathLike


class DOSCAR(LobsterFile):
    """Represents LOBSTER's projected DOS and local projected DOS.

    This class parses and stores data from the DOSCAR file generated by LOBSTER,
    which contains information about the total and projected density of states
    (DOS) for a quantum-chemical calculation performed with VASP.

    Attributes:
        completedos (LobsterCompleteDos): Complete DOS data.
        pdos (list[dict[str, dict[Spin, np.ndarray]]]): Projected DOS data.
            Access as `pdos[atomindex]['orbitalstring'][Spin.up/Spin.down]`.
        tdos (Dos): Total density of states.
        energies (np.ndarray): Energies at which the DOS was calculated
            (in eV, relative to Efermi).
        tdensities (dict[Spin, np.ndarray]): Total density arrays for each spin channel.
        itdensities (dict[Spin, np.ndarray]): Integrated total density arrays for each spin channel.
        is_spin_polarized (bool): Whether the system is spin polarized.
    """

    is_lcfo: ClassVar[bool] = False

    def __init__(
        self,
        filename: PathLike | None = None,
        process_immediately: bool = True,
        lobster_version: str | None = None,
    ) -> None:
        """Initialize a DOSCAR object.

        Args:
            filename (PathLike | None): Path to the DOSCAR file, typically "DOSCAR.lobster".
            process_immediately (bool): Whether to process the file immediately upon initialization.
            lobster_version (str | None): LOBSTER version string to use for parsing. If None, attempts to detect

        Raises:
            ValueError: If neither `structure_file` nor `structure` is provided.
        """
        self.projected_dos: dict[str, dict[str, Dos]]
        self.total_dos: Dos
        self.integrated_total_dos: Dos

        super().__init__(
            filename=filename,
            process_immediately=process_immediately,
            lobster_version=lobster_version,
        )

    @version_processor()
    def process(self) -> None:
        """Process the DOSCAR file and extract DOS data.

        Parses the DOSCAR file to extract total DOS, projected DOS, energies,
        and integrated densities. Sets the appropriate attributes based on
        whether the calculation is spin-polarized.

        Raises:
            ValueError: If the DOSCAR file format is invalid or spin polarization
                cannot be determined.
        """
        total_dos, integrated_total_dos = {}, {}
        data: list[NDArray] = []

        centers: list[str] = []
        orbitals: list[list[str]] = []

        self.spins = [Spin.up]

        header_regex = r"\s*\S+\s+\S+\s+(\d+)\s+(\S+)\s+1\.0+(?:;(.*);(.*))?"

        efermi = None
        ndos = 0

        lines_iter = iter(self.iterate_lines())

        center_counts = defaultdict(int)

        for line in islice(lines_iter, 5, None):
            if match := re.match(header_regex, line):
                ndos = int(match.group(1))

                if efermi is None:
                    efermi = float(match.group(2))

                if center_match := match.group(3):
                    center_match = center_match.strip()

                    if center_match.startswith("Z="):
                        center_match = Element.from_Z(int(center_match.split()[-1])).symbol

                    center_counts[center_match] += 1

                    separator = "_" if self.is_lcfo else ""

                    centers.append(f"{center_match}{separator}{center_counts[center_match]}")

                if orbital_match := match.group(4):
                    orbitals += [[orb.strip() for orb in orbital_match.split()]]

            tmp_dos = []
            if line.strip():
                for _ in range(ndos):
                    line = next(lines_iter).split()
                    tmp_dos.append(line)

                data.append(np.array(tmp_dos, dtype=float))

        if len(data[0][0, :]) == 5:
            self.spins.append(Spin.down)
        elif len(data[0][0, :]) != 3:
            raise ValueError("There is something wrong with the DOSCAR. Can't extract spin polarization.")

        if efermi is None:
            raise ValueError("There is something wrong with the DOSCAR. Can't find efermi.")

        energies = data[0][:, 0]
        projected_dos = {}

        if self.is_spin_polarized:
            total_dos[Spin.up] = data[0][:, 1]
            total_dos[Spin.down] = data[0][:, 2]

            integrated_total_dos[Spin.up] = data[0][:, 3]
            integrated_total_dos[Spin.down] = data[0][:, 4]
        else:
            total_dos[Spin.up] = data[0][:, 1]
            integrated_total_dos[Spin.up] = data[0][:, 2]

        for atom_counter in range(len(data) - 1):
            block_data = data[atom_counter + 1]
            center = centers[atom_counter]

            if center not in projected_dos:
                projected_dos[center] = {}

            for spin_index, spin in enumerate(self.spins):
                for orbital_index, row in enumerate(range(spin_index + 1, block_data.shape[1], len(self.spins))):
                    orbital = orbitals[atom_counter][orbital_index]

                    if orbital not in projected_dos[center]:
                        projected_dos[center][orbital] = {}

                    projected_dos[center][orbital][spin] = block_data[:, row]

        for center, orbitals in projected_dos.items():
            for orbital, dos in orbitals.items():
                projected_dos[center][orbital] = Dos(efermi, energies, dos)

        self.projected_dos: dict[str, dict[str, Dos]] = dict(projected_dos)
        self.total_dos: Dos = Dos(efermi, energies, total_dos)
        self.integrated_total_dos: Dos = Dos(efermi, energies, integrated_total_dos)

    @property
    def efermi(self) -> float:
        """Fermi energy in eV."""
        return self.total_dos.efermi

    @property
    def energies(self) -> NDArray[floating]:
        """Energies at which the DOS was calculated (in eV, relative to efermi)."""
        return self.total_dos.energies

    @property
    def is_spin_polarized(self) -> bool:
        """Whether the system is spin polarized.

        Returns:
            bool: True if the system is spin polarized, False otherwise.
        """
        return len(self.spins) == 2

    @classmethod
    def get_default_filename(cls) -> str:
        """Get the default filename for the DOSCAR.

        Returns:
            str: Default filename string. "DOSCAR.lobster" for regular DOSCAR,
            "DOSCAR.LCFO.lobster" for LCFO analysis.
        """
        return "DOSCAR.lobster" if not cls.is_lcfo else "DOSCAR.LCFO.lobster"


class DOSCAR_LCFO(DOSCAR):
    """Represents LOBSTER's projected DOS and local projected DOS for LCFO analysis.

    This class handles DOSCAR files generated from LCFO analysis, which have
    a different format than regular DOSCAR files.

    Attributes:
        is_lcfo (bool): Class variable indicating this is for LCFO analysis. Always True.
    """

    is_lcfo: ClassVar[bool] = True



"""Module for defining common data used and produced by atomistic simulation packages."""

from __future__ import annotations

import collections
import collections.abc
import importlib
import itertools
import os
import warnings
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, cast

import numpy as np
import orjson
from monty.io import zopen
from monty.json import MSONable
from scipy.interpolate import RegularGridInterpolator

from pymatgen.core import Element, Site, Structure
from pymatgen.core.units import ang_to_bohr, bohr_to_angstrom
from pymatgen.electronic_structure.core import Spin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any, ClassVar, Self, TextIO

    from numpy.typing import ArrayLike, NDArray

    from pymatgen.core.structure import IStructure
    from pymatgen.util.typing import PathLike


class VolumetricData(MSONable):
    """
    A representation of volumetric data commonly used in atomistic simulation outputs,
    such as LOCPOT/CHGCAR files from VASP or cube files from other codes.

    Attributes:
        structure (Structure):
            The crystal structure associated with the volumetric data.
            Represents the lattice and atomic coordinates using the `Structure` class.

        is_spin_polarized (bool):
            Indicates if the simulation is spin-polarized. True for spin-polarized data
            (contains both total and spin-difference densities), False otherwise.

        dim (tuple[int, int, int]):
            The dimensions of the 3D volumetric grid along each axis in the format
            (nx, ny, nz), where nx, ny, and nz represent the number of grid points
            in the x, y, and z directions, respectively.

        data (dict[str, np.ndarray]):
            A dictionary containing the volumetric data. Keys include:
            - `"total"`: A 3D NumPy array representing the total spin density.
            - `"diff"` (optional): A 3D NumPy array representing the spin-difference
              density (spin up - spin down). Typically present in spin-polarized calculations.

        ngridpts (int):
            The total number of grid points in the volumetric data, calculated as
            `nx * ny * nz` using the grid dimensions.
    """

    def __init__(
        self,
        structure: Structure | IStructure,
        data: dict[str, NDArray],
        distance_matrix: dict | None = None,
        data_aug: dict[str, NDArray] | None = None,
    ) -> None:
        """
        Typically, this constructor is not used directly and the static
        from_file constructor is used. This constructor is designed to allow
        summation and other operations between VolumetricData objects.

        Args:
            structure (Structure): associated with the volumetric data
            data (dict[str, NDArray]): Actual volumetric data.
            distance_matrix (NDArray): A pre-computed distance matrix if available.
                Useful so pass distance_matrices between sums,
                short-circuiting an otherwise expensive operation.
            data_aug (NDArray): Any extra information associated with volumetric data
                (typically augmentation charges)
        """
        self.structure = structure
        self.is_spin_polarized: bool = len(data) >= 2
        self.is_soc: bool = len(data) >= 4
        # convert data to numpy arrays in case they were jsanitized as lists
        self.data: dict[str, NDArray] = {k: np.asarray(v) for k, v in data.items()}
        self.dim = self.data["total"].shape
        self.data_aug = data_aug or {}
        self.ngridpts = self.dim[0] * self.dim[1] * self.dim[2]
        # lazy init the spin data since this is not always needed.
        self._spin_data: dict[Spin, NDArray] = {}
        self._distance_matrix = distance_matrix if distance_matrix is not None else {}
        self.xpoints = np.linspace(0.0, 1.0, num=self.dim[0])
        self.ypoints = np.linspace(0.0, 1.0, num=self.dim[1])
        self.zpoints = np.linspace(0.0, 1.0, num=self.dim[2])
        self.interpolator = RegularGridInterpolator(
            (self.xpoints, self.ypoints, self.zpoints),
            self.data["total"],
            bounds_error=True,
        )
        self.name = "VolumetricData"

    def __add__(self, other) -> Self:
        return self.linear_add(other, 1.0)

    def __radd__(self, other) -> Self:
        if other == 0 or other is None:
            # sum() calls 0 + self first; we treat 0 as the identity element
            return self
        if isinstance(other, self.__class__):
            return self.__add__(other)

        raise TypeError(f"Unsupported operand type(s) for +: '{type(other).__name__}' and '{type(self).__name__}'")

    def __sub__(self, other) -> Self:
        return self.linear_add(other, -1.0)

    @property
    def spin_data(self) -> dict[Spin, NDArray]:
        """The data decomposed into actual spin data as {spin: data}.
        Essentially, this provides the actual Spin.up and Spin.down data
        instead of the total and diff. Note that by definition, a
        non-spin-polarized run would have Spin.up data == Spin.down data.
        """
        if not self._spin_data:
            spin_data = {}
            spin_data[Spin.up] = 0.5 * (self.data["total"] + self.data.get("diff", 0))
            spin_data[Spin.down] = 0.5 * (self.data["total"] - self.data.get("diff", 0))
            self._spin_data = spin_data
        return self._spin_data

    def get_axis_grid(self, ind: int) -> list[float]:
        """Get the grid for a particular axis.

        Args:
            ind (int): Axis index.
        """
        ng = self.dim
        num_pts = ng[ind]
        lengths = self.structure.lattice.abc
        return [i / num_pts * lengths[ind] for i in range(num_pts)]

    def copy(self) -> Self:
        """Make a copy of VolumetricData object."""
        return type(self)(
            self.structure,
            {k: v.copy() for k, v in self.data.items()},
            distance_matrix=self._distance_matrix,
            data_aug=self.data_aug,
        )

    def linear_add(self, other, scale_factor: float = 1.0) -> Self:
        """
        Method to do a linear sum of volumetric objects. Used by + and -
        operators as well. Returns a VolumetricData object containing the
        linear sum.

        Args:
            other (VolumetricData): Another VolumetricData object
            scale_factor (float): Factor to scale the other data by.

        Returns:
            VolumetricData corresponding to self + scale_factor * other.
        """
        if self.structure != other.structure:
            warnings.warn(
                "Structures are different. Make sure you know what you are doing...",
                stacklevel=2,
            )
        if list(self.data) != list(other.data):
            raise ValueError("Data have different keys! Maybe one is spin-polarized and the other is not?")

        # To add checks
        data = {}
        for k in self.data:
            data[k] = self.data[k] + scale_factor * other.data[k]

        new = deepcopy(self)
        new.data = data
        new.data_aug = {}
        return new

    def scale(self, factor: float) -> None:
        """Scale the data in place by a factor."""
        for k in self.data:
            self.data[k] = np.multiply(self.data[k], factor)

    def value_at(self, x: float, y: float, z: float) -> float:
        """Get a data value from self.data at a given point (x, y, z) in terms
        of fractional lattice parameters. Will be interpolated using a
        RegularGridInterpolator on self.data if (x, y, z) is not in the original
        set of data points.

        Args:
            x (float): Fraction of lattice vector a.
            y (float): Fraction of lattice vector b.
            z (float): Fraction of lattice vector c.

        Returns:
            Value from self.data (potentially interpolated) corresponding to
            the point (x, y, z).
        """
        return self.interpolator([x, y, z])[0]

    def linear_slice(self, p1: ArrayLike, p2: ArrayLike, n=100):
        """Get a linear slice of the volumetric data with n data points from
        point p1 to point p2, in the form of a list.

        Args:
            p1 (list): 3-element list containing fractional coordinates of the first point.
            p2 (list): 3-element list containing fractional coordinates of the second point.
            n (int): Number of data points to collect, defaults to 100.

        Returns:
            List of n data points (mostly interpolated) representing a linear slice of the
            data from point p1 to point p2.
        """
        p1 = np.asarray(p1)
        p2 = np.asarray(p2)

        if p1.shape != (3,) or p2.shape != (3,):
            raise ValueError(f"lengths of p1 and p2 should be 3, got {len(p1)} and {len(p2)}")

        x_pts = np.linspace(p1[0], p2[0], num=n)
        y_pts = np.linspace(p1[1], p2[1], num=n)
        z_pts = np.linspace(p1[2], p2[2], num=n)
        return [self.value_at(x_pts[i], y_pts[i], z_pts[i]) for i in range(n)]

    def get_integrated_diff(self, ind: int, radius: float, nbins: int = 1) -> NDArray:
        """Get integrated difference of atom index ind up to radius. This can be
        an extremely computationally intensive process, depending on how many
        grid points are in the VolumetricData.

        Args:
            ind (int): Index of atom.
            radius (float): Radius of integration.
            nbins (int): Number of bins. Defaults to 1. This allows one to
                obtain the charge integration up to a list of the cumulative
                charge integration values for radii for [radius/nbins,
                2 * radius/nbins, ....].

        Returns:
            Differential integrated charge as a np array of [[radius, value],
            ...]. Format is for ease of plotting. e.g. plt.plot(data[:,0],
            data[:,1])
        """
        # For non-spin-polarized runs, this is zero by definition.
        if not self.is_spin_polarized:
            radii = [radius / nbins * (i + 1) for i in range(nbins)]
            data = np.zeros((nbins, 2))
            data[:, 0] = radii
            return data

        struct = self.structure
        a = self.dim
        self._distance_matrix = {} if self._distance_matrix is None else self._distance_matrix
        if ind not in self._distance_matrix or self._distance_matrix[ind]["max_radius"] < radius:
            coords = []
            for x, y, z in itertools.product(*(list(range(i)) for i in a)):
                coords.append([x / a[0], y / a[1], z / a[2]])
            sites_dist = struct.lattice.get_points_in_sphere(coords, struct[ind].coords, radius)
            self._distance_matrix[ind] = {
                "max_radius": radius,
                "data": np.array(sites_dist, dtype=object),
            }

        data = self._distance_matrix[ind]["data"]

        # Use boolean indexing to find all charges within the desired distance.
        inds = data[:, 1] <= radius
        dists = data[inds, 1]
        data_inds = np.rint(np.mod(list(data[inds, 0]), 1) * np.tile(a, (len(dists), 1))).astype(int)
        vals = [self.data["diff"][x, y, z] for x, y, z in data_inds]

        hist, edges = np.histogram(dists, bins=nbins, range=(0, radius), weights=vals)
        data = np.zeros((nbins, 2))
        data[:, 0] = edges[1:]
        data[:, 1] = [sum(hist[0 : i + 1]) / self.ngridpts for i in range(nbins)]
        return data

    def get_average_along_axis(self, ind: int) -> NDArray:
        """Get the averaged total of the volumetric data a certain axis direction.
        For example, useful for visualizing Hartree Potentials from a LOCPOT
        file.

        Args:
            ind (int): Index of axis.

        Returns:
            Average total along axis
        """
        total_spin_dens = self.data["total"]
        ng = self.dim
        if ind == 0:
            total = np.sum(np.sum(total_spin_dens, axis=1), 1)
        elif ind == 1:
            total = np.sum(np.sum(total_spin_dens, axis=0), 1)
        else:
            total = np.sum(np.sum(total_spin_dens, axis=0), 0)
        return total / ng[(ind + 1) % 3] / ng[(ind + 2) % 3]

    def to_hdf5(self, filename: PathLike) -> None:
        """Write the VolumetricData to a HDF5 format, which is a highly optimized
        format for reading storing large data. The mapping of the VolumetricData
        to this file format is as follows:

        VolumetricData.data -> f["vdata"]
        VolumetricData.structure ->
            f["Z"]: Sequence of atomic numbers
            f["fcoords"]: Fractional coords
            f["lattice"]: Lattice in the pymatgen.core.Lattice matrix
                format
            f.attrs["structure_json"]: String of JSON representation

        Args:
            filename (str): Filename to output to.
        """
        import h5py

        with h5py.File(str(filename), mode="w") as file:
            ds = file.create_dataset("lattice", (3, 3), dtype="float")
            ds[...] = self.structure.lattice.matrix
            ds = file.create_dataset("Z", (len(self.structure.species),), dtype="i")
            ds[...] = np.array([sp.Z for sp in self.structure.species])
            ds = file.create_dataset("fcoords", self.structure.frac_coords.shape, dtype="float")
            ds[...] = self.structure.frac_coords
            dt = h5py.special_dtype(vlen=str)
            ds = file.create_dataset("species", (len(self.structure.species),), dtype=dt)
            ds[...] = [str(sp) for sp in self.structure.species]
            grp = file.create_group("vdata")
            for k in self.data:
                ds = grp.create_dataset(k, self.data[k].shape, dtype="float")
                ds[...] = self.data[k]
            file.attrs["name"] = self.name
            file.attrs["structure_json"] = orjson.dumps(self.structure.as_dict()).decode()

    @classmethod
    def from_hdf5(cls, filename: PathLike, **kwargs) -> Self:
        """
        Reads VolumetricData from HDF5 file.

        Args:
            filename: Filename

        Returns:
            VolumetricData
        """
        import h5py

        with h5py.File(str(filename), mode="r") as file:
            data = {k: np.asarray(v) for k, v in file["vdata"].items()}
            data_aug = None
            if "vdata_aug" in file:
                data_aug = {k: np.asarray(v) for k, v in file["vdata_aug"].items()}
            structure = Structure.from_dict(orjson.loads(file.attrs["structure_json"]))
            return cls(structure, data=data, data_aug=data_aug, **kwargs)  # type:ignore[arg-type]

    def to_cube(self, filename: PathLike, comment: str = "") -> None:
        """Write the total volumetric data to a cube file format, which consists of two comment lines,
        a header section defining the structure IN BOHR, and the data.

        Args:
            filename (str): Name of the cube file to be written.
            comment (str): If provided, this will be added to the second comment line
        """
        with zopen(filename, mode="wt", encoding="utf-8") as file:
            file = cast("TextIO", file)
            file.write(f"# Cube file for {self.structure.formula} generated by Pymatgen\n")
            file.write(f"# {comment}\n")
            file.write(f"\t {len(self.structure)} 0.000000 0.000000 0.000000\n")

            for idx in range(3):
                lattice_matrix = self.structure.lattice.matrix[idx] / self.dim[idx] * ang_to_bohr
                file.write(
                    f"\t {self.dim[idx]} {lattice_matrix[0]:.6f} {lattice_matrix[1]:.6f} {lattice_matrix[2]:.6f}\n"
                )

            for site in self.structure:
                file.write(
                    f"\t {Element(site.species_string).Z} 0.000000 "
                    f"{ang_to_bohr * site.coords[0]} "
                    f"{ang_to_bohr * site.coords[1]} "
                    f"{ang_to_bohr * site.coords[2]} \n"
                )

            for idx, dat in enumerate(self.data["total"].flatten(), start=1):
                file.write(f"{' ' if dat > 0 else ''}{dat:.6e} ")
                if idx % 6 == 0:
                    file.write("\n")

    @classmethod
    def from_cube(cls, filename: PathLike) -> Self:
        """
        Initialize the cube object and store the data as pymatgen objects.

        Args:
            filename (str): of the cube to read
        """
        # Limit the number of I/O operations by reading the entire file at once
        with zopen(filename, mode="rt", encoding="utf-8") as f:
            lines = f.readlines()

        # Parse the number of atoms from line 3 (first two lines are headers)
        n_atoms = int(lines[2].split()[0])

        # Pre-parse voxel data into arrays
        def parse_voxel(line):
            parts = line.split()
            return int(parts[0]), np.array(parts[1:], dtype=float) * bohr_to_angstrom

        # Extract voxel information from lines 4-6
        num_x_voxels, voxel_x = parse_voxel(lines[3])
        num_y_voxels, voxel_y = parse_voxel(lines[4])
        num_z_voxels, voxel_z = parse_voxel(lines[5])

        # Parse atomic site data (starts at line 7 and continues for n_atoms lines)
        # Each line contains: atomic number, charge, x, y, z coordinates (in Bohr units)
        atom_data_start = 6
        atom_data_end = atom_data_start + n_atoms
        sites = [
            Site(line.split()[0], np.array(line.split()[2:], dtype=float) * bohr_to_angstrom)  # type:ignore[arg-type]
            for line in lines[atom_data_start:atom_data_end]
        ]

        # Generate a Pymatgen Structure object from extracted data
        structure = Structure(
            lattice=[voxel_x * num_x_voxels, voxel_y * num_y_voxels, voxel_z * num_z_voxels],
            species=[site.specie for site in sites],
            coords=[site.coords for site in sites],
            coords_are_cartesian=True,
        )

        # Extract volumetric data (starts after atomic site data)
        volumetric_data_start = atom_data_end
        volumetric_data_lines = lines[volumetric_data_start:]

        # Convert the text-based volumetric data into a NumPy array (much faster)
        volumetric_data = np.fromstring(" ".join(volumetric_data_lines), sep=" ", dtype=float)  # type:ignore[arg-type, type-var]

        # Reshape 1D data array into a 3D grid matching the voxel dimensions
        data = volumetric_data.reshape((num_x_voxels, num_y_voxels, num_z_voxels))

        return cls(structure=structure, data={"total": data})


class PMGDir(collections.abc.Mapping):
    """
    User-friendly class to access all files in a directory as pymatgen objects in a dict. For now, only VASP files are
    implemented but there is no reason why this cannot be extended to other types of files.
    Note that the files are lazily parsed to minimize initialization costs since not all files will be needed by all
    users.

    Example:

    ```python
    pmg_dict = PMGDir(".")
    print(pmg_dict["INCAR"]["NELM"])
    print(pmg_dict["vasprun.xml"].parameters)
    ```
    """

    FILE_MAPPINGS: ClassVar[dict[str, str]] = {
        name: f"pymatgen.io.vasp.{name.capitalize()}"
        for name in (
            "INCAR",
            "POSCAR",
            "KPOINTS",
            "POTCAR",
            "vasprun",
            "OUTCAR",
            "OSZICAR",
            "CHGCAR",
            "WAVECAR",
            "WAVEDER",
            "LOCPOT",
            "XDATCAR",
            "EIGENVAL",
            "PROCAR",
            "ELFCAR",
            "DYNMAT",
        )
    } | {
        "CONTCAR": "pymatgen.io.vasp.Poscar",
        "IBZKPT": "pymatgen.io.vasp.Kpoints",
        "WSWQ": "pymatgen.io.vasp.WSWQ",
    }

    def __init__(self, dirname: PathLike) -> None:
        """
        Args:
            dirname: The directory containing the VASP calculation as a string or Path.
        """
        self.path = Path(dirname).absolute()
        self.reset()

    def __contains__(self, item) -> bool:
        return item in self._files

    def __len__(self) -> int:
        return len(self._files)

    def __iter__(self) -> Iterator[str]:
        return iter(self._files)

    def __getitem__(self, item: str) -> Any:
        if self._files.get(item):
            return self._files.get(item)
        fpath = self.path / item

        if not (self.path / item).exists():
            raise ValueError(f"{item} not found in {self.path}. List of files are {self._files.keys()}.")

        for k, cls_ in PMGDir.FILE_MAPPINGS.items():
            if k in item:
                modname, classname = cls_.rsplit(".", 1)
                module = importlib.import_module(modname)
                class_ = getattr(module, classname)
                try:
                    self._files[item] = class_.from_file(fpath)
                except AttributeError:
                    self._files[item] = class_(fpath)

                return self._files[item]

        warnings.warn(
            f"No parser defined for {item}. Contents are returned as a string.",
            stacklevel=2,
        )
        with zopen(fpath, mode="rt", encoding="utf-8") as f:
            return f.read()

    def __repr__(self) -> str:
        return f"PMGDir({self.path})"

    def reset(self) -> None:
        """
        Reset all loaded files and recheck the directory for files. Use this when the contents of the directory has
        changed.
        """
        # Note that py3.12 has Path.walk(). But we need to use os.walk to ensure backwards compatibility for now.
        self._files: dict[str, Any] = {
            str((Path(d) / f).relative_to(self.path)): None for d, _, fnames in os.walk(self.path) for f in fnames
        }

    def get_files_by_name(self, name: str) -> dict[str, Any]:
        """
        Returns all files with a given name. E.g., if you want all the OUTCAR files, set name="OUTCAR".

        Returns:
            {filename: object from PMGDir[filename]}
        """
        return {f: self[f] for f in self._files if name in f}



"""Classes for reading/manipulating/writing JDFTx input files.

Classes for reading/manipulating/writing JDFTx input files.

Note: JDFTXInfile will be moved back to its own module once a more broad inputs
class is written.
"""

from __future__ import annotations

import logging
import warnings
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, cast

import numpy as np
import scipy.constants as const
from monty.json import MSONable

from pymatgen.core import Lattice, Structure
from pymatgen.core.periodic_table import Element
from pymatgen.core.units import bohr_to_ang
from pymatgen.io.jdftx.generic_tags import AbstractTag, BoolTagContainer, DumpTagContainer, MultiformatTag, TagContainer
from pymatgen.io.jdftx.jdftxinfile_default_inputs import default_inputs
from pymatgen.io.jdftx.jdftxinfile_master_format import (
    __PHONON_TAGS__,
    __TAG_LIST__,
    __WANNIER_TAGS__,
    MASTER_TAG_LIST,
    get_tag_object,
    get_tag_object_on_val,
)
from pymatgen.util.io_utils import clean_lines
from pymatgen.util.typing import SpeciesLike

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Self

    from numpy.typing import ArrayLike

    from pymatgen.util.typing import PathLike

__author__ = "Jacob Clary, Ben Rich"

logger = logging.getLogger(__name__)

# TODO: Add check for whether all ions have or lack velocities.
# TODO: Add default value filling like JDFTx does.
# TODO: Add more robust checking for if two repeatable tag values represent the
# same information. This is likely fixed by implementing filling of default values.
# TODO: Incorporate something to collapse repeated dump tags of the same frequency
# into a single value.


class JDFTXInfile(dict, MSONable):
    """Class for reading/writing JDFtx input files.

    JDFTxInfile object for reading and writing JDFTx input files.
    Essentially a dictionary with some helper functions.
    """

    path_parent: str | None = None  # Only gets a value if JDFTXInfile is initialized with from_file

    def __init__(self, params: dict[str, Any] | None = None) -> None:
        """
        Create a JDFTXInfile object.

        Args:
            params (dict): Input parameters as a dictionary.
        """
        super().__init__()
        if params is not None:
            self.update(params)

    def __str__(self) -> str:
        """Return str representation of JDFTXInfile.

        Returns:
            str: String representation of JDFTXInfile.
        """
        return "".join([line + "\n" for line in self.get_text_list()])

    def __add__(self, other: JDFTXInfile) -> JDFTXInfile:
        """Add existing JDFTXInfile object to method caller JDFTXInfile object.

        Add all the values of another JDFTXInfile object to this object. Facilitate the use of "standard" JDFTXInfiles.
        Repeatable tags are appended together. Non-repeatable tags are replaced with their value from the other object.

        Args:
            other (JDFTXInfile): JDFTXInfile object to add to the method caller object.

        Returns:
            JDFTXInfile: The combined JDFTXInfile object.
        """
        # Deepcopy needed here, or else in `jif1 = jif2 + jif3`, `jif1` will become a reference to `jif2`
        params: dict[str, Any] = deepcopy(dict(self.items()))
        for key, val in other.items():
            if key in self:
                if val is params[key]:
                    # Unlinking the two objects fully cannot be done by deepcopy for some reason
                    continue
                tag_object = get_tag_object(key)
                if tag_object.can_repeat:
                    if isinstance(val, list):
                        for subval in list(val):
                            # Minimum effort to avoid duplicates
                            if subval not in params[key]:
                                params[key].append(subval)
                    else:
                        params[key].append(val)
                else:
                    params[key] = val
            else:
                params[key] = val
        return type(self)(params)

    def __setitem__(self, key: str, value: Any) -> None:
        """Set an item in the JDFTXInfile.

        Set an item in the JDFTXInfile. This magic method is set explicitly to immediately validate when a user sets a
        tag's value, and to perform any conversion necessary.

        Args:
            key (str): Key to set.
            value (Any): Value to set.
        """
        if not isinstance(key, str):
            raise TypeError(f"{key} is not a string!")
        try:
            # tag_object = get_tag_object_on_val(key, value)
            tag_object = get_tag_object(key)
        except KeyError:
            raise KeyError(f"The {key} tag is not in MASTER_TAG_LIST")
        _remaining_values = None
        if tag_object.can_repeat and isinstance(value, list):
            if len(value) > 1:
                _remaining_values = value[1:]
            value = value[0]
        tag_object = get_tag_object_on_val(key, value)

        # if isinstance(tag_object, MultiformatTag):
        #     if isinstance(value, str):
        #         i = tag_object.get_format_index_for_str_value(key, value)
        #     else:
        #         i, _ = tag_object._determine_format_option(key, value)
        #     tag_object = tag_object.format_options[i]
        if tag_object.can_repeat and key in self:
            del self[key]
        if tag_object.can_repeat and not isinstance(value, list):
            value = [value]
        params: dict[str, Any] = {}
        if self._is_numeric(value):
            value = str(value)
        if not tag_object.can_repeat:
            value = [value]  # Shortcut to avoid writing a separate block for non-repeatable tags
        for v in value:
            processed_value = tag_object.read(key, v) if isinstance(v, str) else v
            params = self._store_value(params, tag_object, key, processed_value)
            self.update(params)
            self.validate_tags(try_auto_type_fix=True, error_on_failed_fix=True)
        if _remaining_values is not None:
            for v in _remaining_values:
                self.append_tag(key, v)

    @classmethod
    def from_dict(cls, d: dict[str, Any], validate_value_boundaries=True) -> JDFTXInfile:
        """Create JDFTXInfile from a dictionary.

        Args:
            d (dict): Dictionary to create JDFTXInfile from.

        Returns:
            JDFTXInfile: The created JDFTXInfile object.
        """
        instance = cls()
        for k, v in d.items():
            if k not in ("@module", "@class"):
                instance[k] = v
        if validate_value_boundaries:
            instance.validate_boundaries()
        return instance

    @classmethod
    def from_file(
        cls,
        filename: PathLike,
        dont_require_structure: bool = False,
        sort_tags: bool = True,
        assign_path_parent: bool = True,
        validate_value_boundaries: bool = True,
    ) -> JDFTXInfile:
        """Read a JDFTXInfile object from a file.

        Args:
            filename (PathLike): Filename to read from.
            dont_require_structure (bool, optional): Whether to require structure tags. Defaults to False.
            sort_tags (bool, optional): Whether to sort the tags. Defaults to True.
            assign_path_parent (bool, optional): Whether to assign the parent directory of the input file for include
                tags. Defaults to True.

        Returns:
            JDFTXInfile: The created JDFTXInfile object.
        """
        path_parent = None
        if assign_path_parent:
            path_parent = Path(filename).parents[0]
        with open(filename) as file:
            return cls.from_str(
                file.read(),
                dont_require_structure=dont_require_structure,
                sort_tags=sort_tags,
                path_parent=path_parent,
                validate_value_boundaries=validate_value_boundaries,
            )

    @classmethod
    def from_structure(
        cls,
        structure: Structure,
        selective_dynamics: ArrayLike | None = None,
        write_cart_coords: bool = False,
    ) -> JDFTXInfile:
        """Create a JDFTXInfile object from a pymatgen Structure.

        Args:
            structure (Structure): Structure to convert.
            selective_dynamics (ArrayLike, optional): Selective dynamics attribute for each site if available.
                Shape Nx1, by default None.

        Returns:
            JDFTXInfile: The created JDFTXInfile object.
        """
        jdftxstructure = JDFTXStructure(structure, selective_dynamics)
        jdftxstructure.write_cart_coords = write_cart_coords
        return cls.from_jdftxstructure(jdftxstructure)

    @classmethod
    def from_jdftxstructure(
        cls,
        jdftxstructure: JDFTXStructure,
    ) -> JDFTXInfile:
        """Create a JDFTXInfile object from a JDFTXStructure object.

        Args:
            jdftxstructure (JDFTXStructure): JDFTXStructure object to convert.

        Returns:
            JDFTXInfile: The created JDFTXInfile object.
        """
        jstr = jdftxstructure.get_str()
        return cls.from_str(jstr)

    @classmethod
    def from_str(
        cls,
        string: str,
        dont_require_structure: bool = False,
        sort_tags: bool = True,
        path_parent: Path | None = None,
        validate_value_boundaries: bool = True,
    ) -> JDFTXInfile:
        """Read a JDFTXInfile object from a string.

        Args:
            string (str): String to read from.
            dont_require_structure (bool, optional): Whether to require structure tags. Defaults to False.
            sort_tags (bool, optional): Whether to sort the tags. Defaults to True.
            path_parent (Path, optional): Path to the parent directory of the input file for include tags.
                                          Defaults to None.
            validate_value_boundaries (bool, optional): Whether to validate the value boundaries. Defaults to True.

        Returns:
            JDFTXInfile: The created JDFTXInfile object.
        """
        lines: list[str] = list(clean_lines(string.splitlines()))
        lines = cls._gather_tags(lines)

        params: dict[str, Any] = {}
        # process all tag value lines using specified tag formats in MASTER_TAG_LIST
        for line in lines:
            tag_object, tag, value = cls._preprocess_line(line)
            processed_value = tag_object.read(tag, value)
            params = cls._store_value(params, tag_object, tag, processed_value)  # this will change with tag categories

        if "include" in params:
            for filename in params["include"]:
                _filename = filename
                if not Path(_filename).exists():
                    if path_parent is not None:
                        _filename = path_parent / filename
                    if not Path(_filename).exists():
                        raise ValueError(f"The include file {filename} ({_filename}) does not exist!")
                params.update(cls.from_file(_filename, dont_require_structure=True, assign_path_parent=False))
            del params["include"]

        if (
            not dont_require_structure
            and "lattice" not in params
            and "ion" not in params
            and "ion-species" not in params
        ):
            raise ValueError("This input file is missing required structure tags")
        if sort_tags:
            params = {tag: params[tag] for tag in __TAG_LIST__ if tag in params}
        instance = cls(params)
        if validate_value_boundaries:
            instance.validate_boundaries()
        return instance

    @classmethod
    def get_list_representation(cls, jdftxinfile: JDFTXInfile) -> JDFTXInfile:
        """Convert JDFTXInfile object properties into list representation.

        Args:
            jdftxinfile (JDFTXInfile): JDFTXInfile object to convert.
        """
        reformatted_params = deepcopy(jdftxinfile.as_dict(skip_module_keys=True))
        # rest of code assumes lists are lists and not np.arrays
        reformatted_params = {k: v.tolist() if isinstance(v, np.ndarray) else v for k, v in reformatted_params.items()}
        for tag, value in reformatted_params.items():
            tag_object = get_tag_object(tag)
            if all(
                [
                    tag_object.allow_list_representation,
                    tag_object.is_tag_container,
                    cls._needs_conversion("dict-to-list", value),
                ]
            ):
                reformatted_params.update({tag: tag_object.get_list_representation(tag, value)})
        return cls(reformatted_params)

    @classmethod
    def get_dict_representation(cls, jdftxinfile: JDFTXInfile) -> JDFTXInfile:
        """Convert JDFTXInfile object properties into dict representation.

        Args:
            jdftxinfile (JDFTXInfile): JDFTXInfile object to convert.
        """
        reformatted_params = deepcopy(jdftxinfile.as_dict(skip_module_keys=True))
        # Just to make sure only passing lists and no more numpy arrays
        reformatted_params = {
            k: v.tolist() if isinstance(v, np.ndarray) else v for k, v in reformatted_params.items()
        }  # rest of code assumes lists are lists and not np.arrays
        for tag, value in reformatted_params.items():
            tag_object = get_tag_object(tag)
            if all(
                [
                    tag_object.allow_list_representation,
                    tag_object.is_tag_container,
                    cls._needs_conversion("list-to-dict", value),
                ]
            ):
                reformatted_params.update({tag: tag_object.get_dict_representation(tag, value)})
        return cls(reformatted_params)

    def as_dict(self, sort_tags: bool = True, skip_module_keys: bool = False) -> dict:
        """Return JDFTXInfile as MSONable dict.

        Args:
            sort_tags (bool, optional): Whether to sort the tags. Defaults to True.
            skip_module_keys (bool, optional): Whether to skip the module keys. Defaults to False.

        Returns:
            dict: JDFTXInfile as MSONable dict.
        """
        params = dict(self)
        if sort_tags:
            params = {tag: params[tag] for tag in __TAG_LIST__ if tag in params}
        if not skip_module_keys:
            params["@module"] = type(self).__module__
            params["@class"] = type(self).__name__
        return params

    def copy(self) -> JDFTXInfile:
        """Return a copy of the JDFTXInfile object.

        Returns:
            JDFTXInfile: Copy of the JDFTXInfile object.
        """
        return type(self)(self)

    def get_text_list(self) -> list[str]:
        """Get a list of strings representation of the JDFTXInfile.

        Returns:
            list[str]: List of strings representation of the JDFTXInfile.
        """
        self_as_dict = self.get_dict_representation(self)

        text: list[str] = []
        for tag_group in MASTER_TAG_LIST:
            added_tag_in_group = False
            for tag in MASTER_TAG_LIST[tag_group]:
                if tag not in self:
                    continue
                added_tag_in_group = True
                tag_object: AbstractTag = MASTER_TAG_LIST[tag_group][tag]
                if isinstance(tag_object, MultiformatTag):
                    i, _ = tag_object._determine_format_option(tag, self_as_dict[tag])
                    tag_object = tag_object.format_options[i]
                if tag_object.can_repeat and isinstance(self_as_dict[tag], list):
                    text += [tag_object.write(tag, entry) for entry in self_as_dict[tag]]
                else:
                    text.append(tag_object.write(tag, self_as_dict[tag]))
            if added_tag_in_group:
                text.append("")
        return text

    def write_file(self, filename: PathLike) -> None:
        """Write JDFTXInfile to an in file.

        Args:
            filename (PathLike): Filename to write to.
        """
        with open(filename, mode="w") as file:
            file.write(str(self))

    @classmethod
    def to_jdftxstructure(cls, jdftxinfile: JDFTXInfile, sort_structure: bool = False) -> JDFTXStructure:
        """Convert JDFTXInfile to JDFTXStructure object.

        Converts JDFTx lattice, lattice-scale, ion tags into JDFTXStructure, with Pymatgen structure as attribute.

        Args:
            jdftxinfile (JDFTXInfile): JDFTXInfile object to convert.
            sort_structure (bool, optional): Whether to sort the structure. Useful if species are not grouped properly
                together. Defaults to False.
        """
        # use dict representation so it's easy to get the right column for moveScale,
        # rather than checking for velocities
        jdftxinfile_dict = cls.get_dict_representation(jdftxinfile)
        return JDFTXStructure.from_jdftxinfile(jdftxinfile_dict, sort_structure=sort_structure)

    @classmethod
    def to_pmg_structure(cls, jdftxinfile: JDFTXInfile, sort_structure: bool = False) -> Structure | None:
        """Convert JDFTXInfile to pymatgen Structure object.

        Converts JDFTx lattice, lattice-scale, ion tags into pymatgen Structure.

        Args:
            jdftxinfile (JDFTXInfile): JDFTXInfile object to convert.
            sort_structure (bool, optional): Whether to sort the structure. Useful if species are not grouped properly
                together. Defaults to False.

        Returns:
            Structure: The created pymatgen Structure object.
        """
        # use dict representation so it's easy to get the right column for
        # moveScale, rather than checking for velocities
        logger.info(
            "INPUT DICT REP: %s",
            jdftxinfile.get_dict_representation(jdftxinfile),
        )
        jdftxstructure = JDFTXStructure.from_jdftxinfile(
            jdftxinfile.get_dict_representation(jdftxinfile),
            sort_structure=sort_structure,
        )
        return jdftxstructure.structure

    def strip_structure_tags(self) -> None:
        """Strip all structural tags from the JDFTXInfile.

        Strip all structural tags from the JDFTXInfile. This is useful for preparing a JDFTXInfile object
        from a pre-existing calculation for a new structure.
        """
        strucural_tags = ["lattice", "ion", "lattice-scale", "coords-type"]
        for tag in strucural_tags:
            if tag in self:
                del self[tag]

    def append_tag(self, tag: str, value: Any) -> None:
        """Append a value to a tag.

        Append a value to a tag. Use this method instead of directly appending the list contained in the tag, such that
        the value is properly processed.

        Args:
            tag (str): Tag to append to.
            value (Any): Value to append.
        """
        tag_object = get_tag_object(tag)
        if isinstance(tag_object, MultiformatTag):
            if isinstance(value, str):
                i = tag_object.get_format_index_for_str_value(tag, value)
            else:
                i, _ = tag_object._determine_format_option(tag, value)
            tag_object = tag_object.format_options[i]
        if not tag_object.can_repeat:
            raise ValueError(f"The tag '{tag}' cannot be repeated and thus cannot be appended")
        params: dict[str, Any] = self.as_dict(skip_module_keys=True)
        processed_value = tag_object.read(tag, value) if isinstance(value, str) else value
        params = self._store_value(params, tag_object, tag, processed_value)
        self.update(params)

    # This method is called by setitem, but setitem is circumvented by update,
    # so this method's parameters is still necessary
    def validate_tags(
        self,
        try_auto_type_fix: bool = False,
        error_on_failed_fix: bool = True,
        return_list_rep: bool = False,
    ) -> None:
        """Validate the tags in the JDFTXInfile.

        Validate the tags in the JDFTXInfile. If try_auto_type_fix is True, will attempt to fix the tags. If
        error_on_failed_fix is True, will raise an error if the tags cannot be fixed. If return_list_rep is True, will
        return the tags in list representation.

        Args:
            try_auto_type_fix (bool, optional): Whether to attempt to fix the tags. Defaults to False.
            error_on_failed_fix (bool, optional): Whether to raise an error if the tags cannot be fixed.
                                                  Defaults to True.
            return_list_rep (bool, optional): Whether to return the tags in list representation. Defaults to False.
        """
        for tag in self:
            tag_object = get_tag_object(tag)
            _checked_tag, is_tag_valid, value = tag_object.validate_value_type(
                tag, self[tag], try_auto_type_fix=try_auto_type_fix
            )
            should_warn = not is_tag_valid
            if return_list_rep and tag_object.allow_list_representation:
                value = tag_object.get_list_representation(tag, value)
            if error_on_failed_fix and should_warn and try_auto_type_fix:
                raise ValueError(f"The {tag} tag with value:\n{self[tag]}\ncould not be fixed!")
            if try_auto_type_fix and is_tag_valid:
                self.update({tag: value})
            if should_warn:
                warnmsg = f"The {tag} tag with value:\n{self[tag]}\nhas incorrect typing!"
                if any(isinstance(tag_object, tc) for tc in [TagContainer, DumpTagContainer, BoolTagContainer]):
                    warnmsg += "(Check earlier warnings for more details)\n"
                warnings.warn(warnmsg, stacklevel=2)

    def validate_boundaries(self) -> None:
        """Validate the boundaries of the JDFTXInfile.

        Validate the boundaries of the JDFTXInfile. This is a placeholder for future functionality.
        """
        error_strs: list[str] = []
        for tag in self:
            tag_object = get_tag_object(tag)
            is_valid, error_str = tag_object.validate_value_bounds(tag, self[tag])
            if not is_valid:
                error_strs.append(error_str)
        if len(error_strs) > 0:
            err_cat = "\n".join(error_strs)
            raise ValueError(
                f"The following boundary errors were found in the JDFTXInfile:\n{err_cat}\n"
                "\n Hint - if you are reading from a JDFTX out file, you need to set validate_value_boundaries "
                "to False, as JDFTx will dump values at non-inclusive boundaries (ie 0.0 for values strictly > 0.0)."
            )

    # TODO: Add examples in the docustring on how to use this function
    def is_comparable_to(
        self,
        other: JDFTXInfile,
        exclude_tags: list[str] | None = None,
        exclude_tag_categories: list[str] | None = None,
        ensure_include_tags: list[str] | None = None,
    ) -> bool:
        """Check if two JDFTXInfile objects are comparable.

        Args:
            other (JDFTXInfile): Other JDFTXInfile object to compare to.
            exclude_tags (list[str], optional): Tags to exclude from comparison. Defaults to None.
                Set to contain tags you are already accounting for being different to exclude them.
            exclude_tag_categories (list[str], optional): Tag categories to exclude from comparison. Defaults to None.
                If left as None, will exclude all tags in the "export", "restart", and "structure" categories.
                Add all tags of specified tag categories to be excluded in the comparison.
            ensure_include_tags (list[str], optional): Tags to ensure are included in the comparison. Defaults to None.
                If left as None, will make sure 'ion-species' is included in comparison.
                Helpful for tags in categories you want to exclude but want to ensure are included in the comparison.

        Returns:
            bool: Whether the two JDFTXInfile objects are comparable.
        """
        if exclude_tag_categories is None:
            exclude_tag_categories = ["export", "restart", "structure"]
        if ensure_include_tags is None:
            ensure_include_tags = []
        differing_tags = self.get_filtered_differing_tags(
            other,
            exclude_tags=exclude_tags,
            exclude_tag_categories=exclude_tag_categories,
            ensure_include_tags=ensure_include_tags,
        )
        return len(differing_tags) == 0

    def get_filtered_differing_tags(
        self,
        other: JDFTXInfile,
        exclude_tags: list[str] | None = None,
        exclude_tag_categories: list[str] | None = None,
        ensure_include_tags: list[str] | None = None,
    ) -> list[str]:
        """Return which tags differ between self and other JDFTXInfile.

        Returns a list of tags that differ between the two JDFTXInfile objects, excluding any tags specified in
        exclude_tags or exclude_tag_categories. This is useful for comparing two JDFTXInfile objects while ignoring
        certain tags that are expected to differ.

        Args:
            other (JDFTXInfile): Other JDFTXInfile object to compare to.
            exclude_tags (list[str], optional): Tags to exclude from comparison. Defaults to None.
                Set to contain tags you are already accounting for being different to exclude them.
            exclude_tag_categories (list[str], optional): Tag categories to exclude from comparison. Defaults to None.
                Add all tags of specified tag categories to be excluded in the comparison.
            ensure_include_tags (list[str], optional): Tags to ensure are included in the comparison. Defaults to None.
                Helpful for tags in categories you want to exclude but want to ensure are included in the comparison.
        Returns:
            list[str]: Tags that differ between the two JDFTXInfile objects.
        """
        if exclude_tags is None:
            exclude_tags = []
        if exclude_tag_categories is None:
            exclude_tag_categories = []
        if ensure_include_tags is None:
            ensure_include_tags = []
        for category in exclude_tag_categories:
            exclude_tags += list(MASTER_TAG_LIST[category].keys())
        for tag in ensure_include_tags:
            if tag in exclude_tags:
                exclude_tags.remove(tag)
        differing_tags = self.get_differing_tags(other)
        for exclude_tag in exclude_tags:
            if exclude_tag in differing_tags:
                differing_tags.remove(exclude_tag)
        return differing_tags

    # get_differing_tags for a full list of tags that differ between two JDFTXInfile objects
    def get_differing_tags(self, other: JDFTXInfile) -> list[str]:
        """Return which tags differ between self and other JDFTXInfile.

        Args:
            other (JDFTXInfile): Other JDFTXInfile object to compare to.

        Returns:
            list[str]: Tags that differ between the two JDFTXInfile objects.
        """
        list1 = self.get_differing_tags_from(other)
        list2 = other.get_differing_tags_from(self)
        return list(set(list1 + list2))

    # get_differing_tags_from for only tags contained in self that differ from other JDFTXInfile
    def get_differing_tags_from(self, other: JDFTXInfile) -> list[str]:
        """Return which self-contained tags differ from other JDFTXInfile.

        Args:
            other (JDFTXInfile): Other JDFTXInfile object to compare to.

        Returns:
            list[str]: Tags from this JDFTXInfile that differ from the other JDFTXInfile.
        """
        # TODO: `TagContainer.is_equal_to(val1, tc2, val2)`, where val2 is the fetched default
        # value for the tag, will return a false negative if val1 does not specify all the
        # possible values for a tag. To fix this, missing subtags in both vals for tagcontainer
        # comparisons should be populated with the default values for the tag.
        # (ie if val1 = {a: 1}, val2 = {b: 2}, and the default value for the tag is {a: 1, b: 2},
        # then val1 -> {a: 1, b: 2}, val2 -> {a: 1, b: 2} and the comparison will return True, but
        # if val2 = {b: 3}, then val2 -> {a: 1, b: 3} and the comparison will return False)
        differing_tags = []
        for tag in self:
            val1 = self[tag]
            obj1 = get_tag_object_on_val(tag, val1)
            val2 = None
            obj2 = None
            if tag not in other:
                if tag in ref_infile:
                    val2 = ref_infile[tag]
                    obj2 = get_tag_object_on_val(tag, val2)
                else:
                    differing_tags.append(tag)
            else:
                val2 = other[tag]
                obj2 = get_tag_object_on_val(tag, val2)
            if not (val2 is None or obj2 is None) and not obj1.is_equal_to(val1, obj2, val2):
                differing_tags.append(tag)
        return differing_tags

    @property
    def structure(self) -> Structure | None:
        """Return a pymatgen Structure object.

        Returns:
            Structure: Pymatgen structure object.
        """
        return self.to_pmg_structure(self)

    @classmethod
    def _from_dict(cls, dct: dict[str, Any]) -> JDFTXInfile:
        """Parse a dictionary to create a JDFTXInfile object.

        Args:
            dct (dict): Dictionary to parse.

        Returns:
            JDFTXInfile: The created JDFTXInfile object.
        """
        temp = cls({k: v for k, v in dct.items() if k not in ("@module", "@class")})
        temp = cls.get_dict_representation(cls.get_list_representation(temp))
        return cls.get_list_representation(temp)

    @staticmethod
    def _preprocess_line(line: str) -> tuple[AbstractTag, str, str]:
        """Preprocess a line from a JDFTXInfile, splitting it into a tag object, tag name, and value.

        Args:
            line (str): Line from the input file.

        Returns:
            tuple[AbstractTag, str, str]: Tag object, tag name, and value.
        """
        line_list = line.strip().split(maxsplit=1)
        tag: str = line_list[0].strip()
        if tag in __PHONON_TAGS__:
            raise ValueError("Phonon functionality has not been added!")
        if tag in __WANNIER_TAGS__:
            raise ValueError("Wannier functionality has not been added!")
        if tag not in __TAG_LIST__:
            err_str = f"The {tag} tag in {line_list} is not in MASTER_TAG_LIST and is not a comment, "
            err_str += "something is wrong with this input data!"
            raise ValueError(err_str)
        tag_object = get_tag_object(tag)
        value: str = ""
        if len(line_list) == 2:
            value = line_list[1].strip()
        elif len(line_list) == 1:
            value = ""  # exception for tags where only tagname is used, e.g. dump-only tag
        if isinstance(tag_object, MultiformatTag):
            i = tag_object.get_format_index_for_str_value(tag, value)
            tag_object = tag_object.format_options[i]
        return tag_object, tag, value

    @staticmethod
    def _store_value(
        params: dict[str, list | list[dict[str, dict]] | Any],
        tag_object: AbstractTag,
        tag: str,
        value: Any,
    ) -> dict:
        """Store the value in the params dictionary.

        Args:
            params (dict): Dictionary to store the value in.
            tag_object (AbstractTag): Tag object for holding tag/value pair. This tag object should be the tag object
                which the "tag" string maps to in MASTER_TAG_LIST.
            tag (str): Tag name.
            value (Any): Value to store.

        Returns:
            dict: Updated dictionary with the stored value.
        """
        if tag_object.can_repeat:  # store tags that can repeat in a list
            if tag not in params:
                params[tag] = []
            params[tag].append(value)
        else:
            if tag in params:
                raise ValueError(f"The '{tag}' tag appears multiple times in this input when it should not!")
            params[tag] = value
        return params

    @staticmethod
    def _gather_tags(lines: list[str]) -> list[str]:
        """Gather broken lines into single string for processing later.

        Args:
            lines (list[str]): List of lines from the input file.

        Returns:
            list[str]: List of strings with tags broken across lines combined into single string.
        """
        total_tag = ""
        gathered_strings = []
        for line in lines:
            if line[-1] == "\\":  # "\" indicates line continuation (the first "\" needed to be escaped)
                total_tag += line[:-1].strip() + " "  # remove \ and any extra whitespace
            elif total_tag:  # then finished with line continuations
                total_tag += line
                gathered_strings.append(total_tag)
                total_tag = ""
            else:  # then append line like normal
                gathered_strings.append(line)
        return gathered_strings

    @staticmethod
    def _needs_conversion(conversion: str, value: dict | list[dict] | list | list[list]) -> bool:
        """Determine if a value needs to be converted.

        This method is only ever called by cls.get_list/dict_representation.

        Args:
            conversion (str): Conversion type. ('dict-to-list' (value : dict | list[dict]) or
                                'list-to-dict' (value : list | list[list]))
            value (dict | list[dict] | list | list[list]): Value to check.

        Returns:
            bool: Whether the value needs to be converted.
        """
        # Check if value is not iterable
        try:
            iter(value)
        except TypeError:
            # This is triggered when JDFTXInfile is attempting to convert a non-tagcontainer to list/dict representation
            # The return boolean is meaningless in this case, so just returning False to avoid the value hitting the
            # "for x in value" loop below.
            return False
        if conversion == "list-to-dict":
            flag = False
        elif conversion == "dict-to-list":
            flag = True
        else:
            raise ValueError(f"Conversion type {conversion} is not 'list-to-dict' or 'dict-to-list'")
        if isinstance(value, dict) or all(isinstance(x, dict) for x in value):
            return flag
        return not flag

    def _is_numeric(self, value: Any) -> bool:
        """Check if a value is numeric.

        Args:
            value (Any): Value to check.

        Returns:
            bool: Whether the value is numeric.
        """
        # data-types that might accidentally be identified as numeric
        if isinstance(value, bool):
            return False
        try:
            float(value)
            is_numeric = True
        except (ValueError, TypeError):
            is_numeric = False
        return is_numeric


ref_infile = JDFTXInfile.from_dict(default_inputs, validate_value_boundaries=False)


def _allnone(val) -> bool:
    """Check if all elements in a list are None.

    Args:
        val (list | None): List to check.

    Returns:
        bool: Whether all elements in the list are None.
    """
    if val is None:
        return True
    if isinstance(val, list):
        return all(_allnone(x) for x in val)
    return False


def selective_dynamics_site_prop_to_jdftx_interpretable(selective_dynamics: Sequence[Any]) -> ArrayLike:
    """Convert selective dynamics site property to JDFTX interpretable format.

    Args:
        selective_dynamics (list[list[bool]]): Selective dynamics site property.

    Returns:
        ArrayLike: JDFTX interpretable format of selective dynamics.
    """
    _selective_dynamics = []
    for i in range(len(selective_dynamics)):
        # Not worrying about partial free axes for now, just setting movescale to 0 if any axis is fixed
        if any(not x for x in selective_dynamics[i]):
            _selective_dynamics.append(0)
        else:
            _selective_dynamics.append(1)
    return np.array(_selective_dynamics)


@dataclass
class JDFTXStructure(MSONable):
    """Object for representing the data in JDFTXStructure tags.

    Attributes:
        structure (Structure): Associated Structure.
        selective_dynamics (ArrayLike): Selective dynamics attribute for each site if available. Shape Nx1.
        sort_structure (bool): Whether to sort the structure. Useful if species are not grouped properly together.
            Defaults to False.
        write_cart_coords (bool): Whether to write cartesian coordinates. Defaults to False.
        velocities (list[np.ndarray | list[float] | None]): Velocities for each site if available. Kept separate from
            structure.site_properties["velocities"] to allow inhomogeneous shaping.
        constraint_types (list[str | None]): Constraint types for each site if available. Kept separate from
            structure.site_properties["constraint_types"] to allow inhomogeneous shaping.
        constraint_vectors (list[np.ndarray | list[np.ndarray] | None]): Constraint vectors for each site if
            available. Kept separate from structure.site_properties["constraint_vectors"] to allow inhomogeneous
            shaping. i'th entry must be a list of vectors if the i'th constraint_type is "HyperPlane". Otherwise,
            a single vector is expected.
        hyperplane_group_names (list[list[str] | None]): Group names for each site if available. Kept separate from
            structure.site_properties["group_names"] to allow inhomogeneous shaping. i'th entry must be a list of group
            names if the i'th constraint_type is "HyperPlane". Otherwise, None is expected.
    """

    structure: Structure | None = None
    selective_dynamics: ArrayLike | None = None
    sort_structure: bool = False
    write_cart_coords: bool = False
    velocities: list | None = None
    constraint_types: list | None = None
    constraint_vectors: list | None = None
    hyperplane_group_names: list | None = None

    def __post_init__(self) -> None:
        """Post init function for JDFTXStructure.

        Asserts self.structure is ordered, and adds selective dynamics if needed.
        """
        if ((self.structure is not None) and ("selective_dynamics" in self.structure.site_properties)) and (
            self.selective_dynamics is None
        ):
            self.selective_dynamics = selective_dynamics_site_prop_to_jdftx_interpretable(
                self.structure.site_properties["selective_dynamics"]
            )
        for i in range(len(self.structure.species)):
            name = ""
            if isinstance(self.structure.species[i], str):
                _name_str = self.structure.species[i]
            elif isinstance(self.structure.species[i], SpeciesLike):
                _name_str = self.structure.species[i].symbol
                if not isinstance(_name_str, str):
                    _name_str = _name_str.symbol
            else:
                raise TypeError("Species must be a string or SpeciesLike object")
            name_str = str(_name_str)
            for j in range(len(name_str)):
                if not name_str[j].isdigit():
                    name += name_str[j]
            self.structure.species[i] = name
        if self.structure.is_ordered:
            site_properties = {}
            if self.selective_dynamics is not None:
                selective_dynamics = np.array(self.selective_dynamics)
                if not selective_dynamics.all():
                    site_properties["selective_dynamics"] = selective_dynamics

            # create new copy of structure so can add selective dynamics and sort atoms if needed
            structure = Structure.from_sites(self.structure.sites)
            self.structure = structure.copy(site_properties=site_properties)
            if self.sort_structure:
                self.structure = self.structure.get_sorted_structure()
        else:
            raise ValueError("Disordered structure with partial occupancies cannot be converted into JDFTXStructure!")
        if _allnone(self.selective_dynamics):
            self.selective_dynamics = None
        if _allnone(self.velocities):
            self.velocities = None
        if _allnone(self.constraint_types):
            self.constraint_types = None
        if _allnone(self.constraint_vectors):
            self.constraint_vectors = None
        if _allnone(self.hyperplane_group_names):
            self.hyperplane_group_names = None

    def __repr__(self) -> str:
        """Return representation of JDFTXStructure file.

        Returns:
            str: Representation of JDFTXStructure file.
        """
        return f"JDFTXStructure({self.get_str()})"

    def __str__(self) -> str:
        """Return string representation of JDFTXStructure file.

        Returns:
            str: String representation of JDFTXStructure file.
        """
        return self.get_str()

    @property
    def natoms(self) -> int:
        """Return number of atoms.

        Returns:
            int: Number of sites.
        """
        return len(self.structure.species)

    @classmethod
    def from_str(cls, data: str) -> JDFTXStructure:
        """Read JDFTXStructure from string.

        Args:
            data (str): String to read from.

        Returns:
            JDFTXStructure: The created JDFTXStructure object.
        """
        return cls.from_jdftxinfile(JDFTXInfile.from_str(data))

    @classmethod
    def from_file(cls, filename: str) -> JDFTXStructure:
        """Read JDFTXStructure from file.

        Args:
            filename (str): Filename to read from.

        Returns:
            JDFTXStructure: The created JDFTXStructure object.
        """
        return cls.from_jdftxinfile(JDFTXInfile.from_file(filename))

    @classmethod
    def from_jdftxinfile(cls, jdftxinfile: JDFTXInfile, sort_structure: bool = False) -> JDFTXStructure:
        """Get JDFTXStructure from JDFTXInfile.

        Args:
            jdftxinfile (JDFTXInfile): JDFTXInfile object.
            sort_structure (bool, optional): Whether to sort the structure. Useful if species are not grouped properly
                together as JDFTx output will have species sorted.

        Returns:
            JDFTXStructure: The created JDFTXStructure object.
        """

        lattice = _infile_to_pmg_lattice(jdftxinfile)
        atomic_symbols = [x["species-id"] for x in jdftxinfile["ion"]]
        coords = np.array([[x["x0"], x["x1"], x["x2"]] for x in jdftxinfile["ion"]])
        coords *= const.value("Bohr radius") * 10**10  # Bohr radius in Ang; convert to Ang
        selective_dynamics = np.array([x["moveScale"] for x in jdftxinfile["ion"]])
        # Gather optional tags
        velocities = [
            np.array([float(v) for v in [x["v"]["vx0"], x["v"]["vx1"], x["v"]["vx2"]]]) if "v" in x else None
            for x in jdftxinfile["ion"]
        ]
        constraint_types = [x.get("constraint-type", None) for x in jdftxinfile["ion"]]
        constraint_vectors = [
            [x["d0"], x["d1"], x["d2"]] if "constraint-type" in x else None for x in jdftxinfile["ion"]
        ]
        hyperplane_vectors = [
            [[[y["d0"], y["d1"], y["d2"]] for y in x["HyperPlane"]]] if "HyperPlane" in x else None
            for x in jdftxinfile["ion"]
        ]
        hyperplane_groups: list = [None for _ in range(len(hyperplane_vectors))]
        if not all(x is None for x in hyperplane_vectors):
            for i in range(len(hyperplane_vectors)):
                if hyperplane_vectors[i] is not None:
                    constraint_types[i] = "HyperPlane"
                    constraint_vectors[i] = hyperplane_vectors[i]
            hyperplane_groups = [
                [[y["group"] for y in x["HyperPlane"]] if "HyperPlane" in x else None] for x in jdftxinfile["ion"]
            ]

        coords_are_cartesian = False  # is default for JDFTx
        if "coords-type" in jdftxinfile:
            coords_are_cartesian = jdftxinfile["coords-type"] == "Cartesian"

        struct: Structure = Structure(
            lattice,
            atomic_symbols,
            coords,
            to_unit_cell=False,
            validate_proximity=False,
            coords_are_cartesian=coords_are_cartesian,
        )
        return cls(
            structure=struct,
            selective_dynamics=selective_dynamics,
            sort_structure=sort_structure,
            velocities=velocities,
            constraint_types=constraint_types,
            constraint_vectors=constraint_vectors,
            hyperplane_group_names=hyperplane_groups,
        )

    def get_str(self, in_cart_coords: bool | None = None) -> str:
        """Return a string to be written as JDFTXInfile tags.

        Allows extra options as compared to calling str(JDFTXStructure) directly.

        Args:
            in_cart_coords (bool, optional): Whether coordinates are output in direct or Cartesian.

        Returns:
            str: Representation of JDFTXInfile structure tags.
        """
        jdftx_tag_dict: dict[str, Any] = {}

        if in_cart_coords is None:
            in_cart_coords = self.write_cart_coords

        lattice = np.copy(self.structure.lattice.matrix)
        lattice = lattice.T  # transpose to get into column-vector format
        lattice /= const.value("Bohr radius") * 10**10  # Bohr radius in Ang; convert to Bohr

        jdftx_tag_dict["lattice"] = {f"R{i}{j}": lattice[i][j] for i in range(3) for j in range(3)}
        jdftx_tag_dict["ion"] = []
        jdftx_tag_dict["coords-type"] = "Cartesian" if in_cart_coords else "Lattice"
        valid_labels = [
            value.symbol for key, value in Element.__dict__.items() if not key.startswith("_") and not callable(value)
        ]
        if (self.constraint_types is not None) and (self.constraint_vectors is None):
            raise ValueError("Constraint types are present but constraint vectors are not!")
        if (self.constraint_vectors is not None) and (self.constraint_types is None):
            raise ValueError("Constraint vectors are present but constraint types are not!")
        if self.structure is None:
            return ""
        for i, site in enumerate(self.structure):
            label = site.label
            coords = site.coords * (1 / bohr_to_ang) if in_cart_coords else site.frac_coords
            sd = int(bool(cast("np.ndarray", self.selective_dynamics)[i])) if self.selective_dynamics is not None else 1
            # TODO: This is needlessly complicated, simplify this
            if label not in valid_labels:
                for varname in ["species_string", "specie.name"]:
                    if _multi_hasattr(site, varname) and _multi_getattr(site, varname) in valid_labels:
                        label = _multi_getattr(site, varname)
                        break
                if label not in valid_labels:
                    raise ValueError(f"Could not correct site label {label} for site (index {i})")
            ion_dict_rep = {
                "species-id": label,
                "x0": coords[0],
                "x1": coords[1],
                "x2": coords[2],
                "moveScale": sd,
            }
            # Gather velocities if present
            if (self.velocities is not None) and (self.velocities[i] is not None):
                ion_dict_rep["v"] = {
                    "vx0": self.velocities[i][0] * (1 / bohr_to_ang),
                    "vx1": self.velocities[i][1] * (1 / bohr_to_ang),
                    "vx2": self.velocities[i][2] * (1 / bohr_to_ang),
                }
            # Gather constraints if present
            if ((self.constraint_types is not None) and (self.constraint_vectors is not None)) and (
                self.constraint_types[i] is not None
            ):
                ctype = self.constraint_types[i]
                if ctype == "HyperPlane":
                    ion_dict_rep["HyperPlane"] = []
                    if self.constraint_vectors[i] is not None:
                        for j, hyperplane in enumerate(self.constraint_vectors[i]):
                            _hyperplane = {}
                            for k in range(3):
                                _hyperplane["d" + str(k)] = hyperplane[k]
                            _hyperplane["group"] = self.hyperplane_groups[i][j]
                            ion_dict_rep["HyperPlane"].append(_hyperplane)
                else:
                    ion_dict_rep["constraint-type"] = ctype
                    for j in range(3):
                        ion_dict_rep["d" + str(j)] = self.constraint_vectors[i][j]
            jdftx_tag_dict["ion"].append(ion_dict_rep)
        return str(JDFTXInfile.from_dict(jdftx_tag_dict))

    def write_file(self, filename: PathLike, **kwargs) -> None:
        """Write JDFTXStructure to a file.

        The supported kwargs are the same as those for the JDFTXStructure.get_str method and are passed through
        directly.

        Args:
            filename (PathLike): Filename to write to.
            **kwargs: Kwargs to pass to JDFTXStructure.get_str.
        """
        with open(filename, mode="w") as file:
            file.write(self.get_str(**kwargs))

    def as_dict(self) -> dict:
        """MSONable dict.

        Returns:
            dict: MSONable dictionary representation of the JDFTXStructure.
        """
        return {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "structure": self.structure.as_dict(),
            "selective_dynamics": np.array(self.selective_dynamics).tolist(),
            "velocities": self.velocities,
            "constraint_types": self.constraint_types,
            "constraint_vectors": self.constraint_vectors,
            "hyperplane_group_names": self.hyperplane_group_names,
        }

    @classmethod
    def from_dict(cls, params: dict) -> Self:
        """Get JDFTXStructure from dict.

        Args:
            params (dict): Serialized JDFTXStructure.

        Returns:
            JDFTXStructure: The created JDFTXStructure object.
        """
        return cls(
            Structure.from_dict(params["structure"]),
            selective_dynamics=params["selective_dynamics"],
        )


def _multi_hasattr(varbase: Any, varname: str):
    """Check if object has an attribute (capable of nesting with . splits).

    Check if object has an attribute (capable of nesting with . splits).

    Parameters
    ----------
    varbase
        Object to check.
    varname
        Attribute to check for.

    Returns
    -------
    bool
        Whether the object has the attribute.
    """
    varlist = varname.split(".")
    for i, var in enumerate(varlist):
        if i == len(varlist) - 1:
            return hasattr(varbase, var)
        if hasattr(varbase, var):
            varbase = getattr(varbase, var)
        else:
            return False
    return None


def _multi_getattr(varbase: Any, varname: str):
    """Check if object has an attribute (capable of nesting with . splits).

    Check if object has an attribute (capable of nesting with . splits).

    Parameters
    ----------
    varbase
        Object to check.
    varname
        Attribute to check for.

    Returns
    -------
    Any
        Attribute of the object.
    """
    if not _multi_hasattr(varbase, varname):
        raise AttributeError(f"{varbase} does not have attribute {varname}")
    varlist = varname.split(".")
    for var in varlist:
        varbase = getattr(varbase, var)
    return varbase


def _infile_to_pmg_lattice(jdftxinfile: JDFTXInfile) -> Lattice:
    jl = jdftxinfile["lattice"]
    if "R00" not in jl:
        return _infile_special_format_to_pmg_lattice(jdftxinfile)
    return _infile_matrix_format_to_pmg_lattice(jdftxinfile)


def _infile_special_format_to_pmg_lattice(jdftxinfile: JDFTXInfile) -> Lattice:
    jl = jdftxinfile["lattice"]
    if "modification" in jl:
        raise NotImplementedError("Special case lattices with modification not implemented yet")
    # Second boolean to check if latt-scale was passed but does nothing
    if ("latt-scale" in jl) and (not all(np.isclose(float(x), 1) in jl for x in ["s0", "s1", "s2"])):
        raise NotImplementedError("latt-scale for special case lattices not implemented yet")
    if "Monoclinic" in jl:
        lattice = Lattice.monoclinic(jl["a"], jl["b"], jl["c"], jl["beta"])
    elif "Orthorhombic" in jl:
        lattice = Lattice.orthorhombic(jl["a"], jl["b"], jl["c"])
    elif "Cubic" in jl:
        lattice = Lattice.cubic(jl["a"])
    elif "Hexagonal" in jl:
        lattice = Lattice.hexagonal(jl["a"], jl["c"])
    elif "Rhombohedral" in jl:
        lattice = Lattice.rhombohedral(jl["a"], jl["alpha"])
    elif "Tetragonal" in jl:
        lattice = Lattice.tetragonal(jl["a"], jl["c"])
    elif "Triclinic" in jl:
        lattice = Lattice.from_parameters(jl["a"], jl["b"], jl["c"], jl["alpha"], jl["beta"], jl["gamma"])
    else:
        raise ValueError(f"Unable to convert JDFTX lattice tag {jl} to pymatgen Lattice object")
    return lattice


def _infile_matrix_format_to_pmg_lattice(jdftxinfile: JDFTXInfile) -> Lattice:
    """Convert JDFTX lattice tag to pymatgen Lattice object.

    Args:
        jl (dict[str, Any]): JDFTX lattice tag.

    Returns:
        Lattice: Pymatgen Lattice object.
    """
    _lattice = np.zeros([3, 3])
    jl = jdftxinfile["lattice"]
    for i in range(3):
        for j in range(3):
            _lattice[i][j] += float(jl[f"R{i}{j}"])
    if "latt-scale" in jdftxinfile:
        latt_scale = np.array([jdftxinfile["latt-scale"][x] for x in ["s0", "s1", "s2"]])
        _lattice *= latt_scale
    _lattice = _lattice.T  # convert to row vector format
    _lattice *= const.value("Bohr radius") * 10**10  # Bohr radius in Ang; convert to Ang
    return Lattice(_lattice)
